from plone import api
from plone.app.testing import setRoles
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import SITE_OWNER_PASSWORD
from plone.app.testing import TEST_USER_ID
from plone.app.testing import TEST_USER_NAME
from plone.app.testing import TEST_USER_PASSWORD
from plone.restapi.testing import PLONE_RESTAPI_DX_FUNCTIONAL_TESTING
from plone.restapi.testing import RelativeSession
from zope.component import getGlobalSiteManager
from zope.component import provideUtility
from zope.componentvocabulary.vocabulary import UtilityTerm
from zope.schema.interfaces import IVocabularyFactory
from zope.schema.vocabulary import SimpleTerm
from zope.schema.vocabulary import SimpleVocabulary

import transaction
import unittest


TEST_TERM_1 = SimpleTerm(42, token="token1", title="Title 1")
TEST_TERM_2 = SimpleTerm(43, token="token2", title="Title 2")
TEST_TERM_3 = SimpleTerm(44, token="token3")
TEST_TERM_4 = UtilityTerm(45, "token4")
TEST_TERM_5 = SimpleTerm(46, token="token5", title="Tötle 5")
TEST_TERM_6 = SimpleTerm(47, token="token6", title="Tötle 6")
TEST_TERM_7 = SimpleTerm(
    48, token="token7", title="This is a title for the seventh term"
)

TEST_VOCABULARY = SimpleVocabulary(
    [
        TEST_TERM_1,
        TEST_TERM_2,
        TEST_TERM_3,
        TEST_TERM_4,
        TEST_TERM_5,
        TEST_TERM_6,
        TEST_TERM_7,
    ]
)


def test_vocabulary_factory(context):
    return TEST_VOCABULARY


TEST_BIG_VOCABULARY = SimpleVocabulary(
    [SimpleTerm(a, token=f"token{a}", title=f"Title {a}") for a in range(100)]
)


def test_big_vocabulary_factory(context):
    return TEST_BIG_VOCABULARY


def test_context_vocabulary_factory(context):
    return SimpleVocabulary(
        [
            SimpleTerm(context.id, token="id", title=context.id),
            SimpleTerm(context.title, token="title", title=context.title),
        ]
    )


class TestVocabularyEndpoint(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_FUNCTIONAL_TESTING

    maxDiff = None

    def setUp(self):
        self.app = self.layer["app"]
        self.portal = self.layer["portal"]
        self.portal_url = self.portal.absolute_url()
        setRoles(self.portal, TEST_USER_ID, ["Manager"])

        self.api_session = RelativeSession(self.portal_url, test=self)
        self.api_session.headers.update({"Accept": "application/json"})
        self.api_session.auth = (SITE_OWNER_NAME, SITE_OWNER_PASSWORD)
        provideUtility(
            test_vocabulary_factory,
            provides=IVocabularyFactory,
            name="plone.restapi.tests.test_vocabulary",
        )

    def test_get_vocabulary(self):
        response = self.api_session.get(
            "/@vocabularies/plone.restapi.tests.test_vocabulary"
        )

        self.assertEqual(200, response.status_code)
        response = response.json()
        self.assertEqual(
            response,
            {
                "@id": self.portal_url
                + "/@vocabularies/plone.restapi.tests.test_vocabulary",  # noqa
                "items": [
                    {"title": "Title 1", "token": "token1"},
                    {"title": "Title 2", "token": "token2"},
                    {"title": "token3", "token": "token3"},
                    {"title": "token4", "token": "token4"},
                    {"title": "T\xf6tle 5", "token": "token5"},
                    {"title": "T\xf6tle 6", "token": "token6"},
                    {
                        "title": "This is a title for the seventh term",
                        "token": "token7",
                    },
                ],
                "items_total": 7,
            },
        )

    def test_get_builtin_vocabulary(self):
        """Check if built-in vocabularies are protected.

        See plone.app.vocabularies.PERMISSIONS
        """
        self.api_session.auth = (TEST_USER_NAME, TEST_USER_PASSWORD)

        # test editor
        setRoles(self.portal, TEST_USER_ID, ["Member", "Contributor", "Editor"])
        transaction.commit()
        response = self.api_session.get(
            "/@vocabularies/plone.app.vocabularies.Keywords"
        )
        self.assertEqual(200, response.status_code)
        response = response.json()
        self.assertEqual(
            response,
            {
                "@id": self.portal_url
                + "/@vocabularies/plone.app.vocabularies.Keywords",  # noqa
                "items": [],
                "items_total": 0,
            },
        )
        # test Anonymous
        setRoles(self.portal, TEST_USER_ID, ["Anonymous"])
        transaction.commit()

        self.api_session.auth = ()
        response = self.api_session.get(
            "/@vocabularies/plone.app.vocabularies.Keywords"
        )
        self.assertEqual(401, response.status_code)

    def test_get_vocabulary_batched(self):
        response = self.api_session.get(
            "/@vocabularies/plone.restapi.tests.test_vocabulary?b_size=1"
        )

        self.assertEqual(200, response.status_code)
        response = response.json()
        self.assertEqual(
            response,
            {
                "@id": self.portal_url
                + "/@vocabularies/plone.restapi.tests.test_vocabulary",  # noqa
                "batching": {
                    "@id": self.portal_url
                    + "/@vocabularies/plone.restapi.tests.test_vocabulary?b_size=1",  # noqa
                    "first": self.portal_url
                    + "/@vocabularies/plone.restapi.tests.test_vocabulary?b_start=0&b_size=1",  # noqa
                    "last": self.portal_url
                    + "/@vocabularies/plone.restapi.tests.test_vocabulary?b_start=6&b_size=1",  # noqa
                    "next": self.portal_url
                    + "/@vocabularies/plone.restapi.tests.test_vocabulary?b_start=1&b_size=1",  # noqa
                },
                "items": [{"title": "Title 1", "token": "token1"}],
                "items_total": 7,
            },
        )

    def test_get_vocabulary_filtered_by_title(self):
        response = self.api_session.get(
            "/@vocabularies/plone.restapi.tests.test_vocabulary?title=This"
        )

        self.assertEqual(200, response.status_code)
        response = response.json()
        self.assertEqual(
            response["items"][0],
            {"title": "This is a title for the seventh term", "token": "token7"},
        )
        self.assertEqual(
            response["items_total"],
            1,
        )

        response = self.api_session.get(
            "/@vocabularies/plone.restapi.tests.test_vocabulary?title=is+a+title"
        )

        self.assertEqual(200, response.status_code)
        response = response.json()
        self.assertEqual(
            response["items"][0],
            {"title": "This is a title for the seventh term", "token": "token7"},
        )

    def test_get_vocabulary_filtered_by_non_ascii_title(self):
        response = self.api_session.get(
            "/@vocabularies/plone.restapi.tests.test_vocabulary?title=t%C3%B6tle"
        )

        self.assertEqual(200, response.status_code)
        response = response.json()
        self.assertEqual(
            response,
            {
                "@id": self.portal_url
                + "/@vocabularies/plone.restapi.tests.test_vocabulary?title=t%C3%B6tle",  # noqa
                "items": [
                    {"title": "T\xf6tle 5", "token": "token5"},
                    {"title": "T\xf6tle 6", "token": "token6"},
                ],
                "items_total": 2,
            },
        )

    def test_get_vocabulary_filtered_by_non_ascii_title_as_plain_utf(self):
        response = self.api_session.get(
            "/@vocabularies/plone.restapi.tests.test_vocabulary?title=tötle"
        )

        self.assertEqual(200, response.status_code)
        response = response.json()
        self.assertEqual(
            response,
            {
                "@id": self.portal_url
                + "/@vocabularies/plone.restapi.tests.test_vocabulary?title=t%C3%B6tle",  # noqa
                "items": [
                    {"title": "Tötle 5", "token": "token5"},
                    {"title": "Tötle 6", "token": "token6"},
                ],
                "items_total": 2,
            },
        )

    def test_get_vocabulary_filtered_by_token(self):
        response = self.api_session.get(
            "/@vocabularies/plone.restapi.tests.test_vocabulary?token=token1"
        )

        self.assertEqual(200, response.status_code)
        response = response.json()
        self.assertEqual(
            response,
            {
                "@id": self.portal_url
                + "/@vocabularies/plone.restapi.tests.test_vocabulary?token=token1",  # noqa
                "items": [{"title": "Title 1", "token": "token1"}],
                "items_total": 1,
            },
        )

    def test_get_vocabulary_filtered_by_token_partial_not_match(self):
        response = self.api_session.get(
            "/@vocabularies/plone.restapi.tests.test_vocabulary?token=token"
        )

        self.assertEqual(200, response.status_code)
        response = response.json()
        self.assertEqual(
            response,
            {
                "@id": self.portal_url
                + "/@vocabularies/plone.restapi.tests.test_vocabulary?token=token",  # noqa
                "items": [],
                "items_total": 0,
            },
        )

    def test_get_vocabulary_filtered_by_title_and_token_returns_error(self):
        response = self.api_session.get(
            "/@vocabularies/plone.restapi.tests.test_vocabulary?token=token1&title=Title"  # noqa
        )

        self.assertEqual(400, response.status_code)
        response = response.json()
        self.assertEqual(
            response.get("error"),
            {
                "message": "You can not filter by title and token at the same time.",  # noqa
                "type": "Invalid parameters",
            },
        )

    def test_get_corner_case_vocabulary_filtered_by_token(self):
        response = self.api_session.get(
            "/@vocabularies/plone.app.vocabularies.Weekdays?token=0"
        )

        self.assertEqual(200, response.status_code)
        response = response.json()
        self.assertEqual(
            response,
            {
                "@id": self.portal_url
                + "/@vocabularies/plone.app.vocabularies.Weekdays?token=0",  # noqa
                "items": [{"title": "Monday", "token": "0"}],
                "items_total": 1,
            },
        )

    def test_get_vocabulary_filtered_by_token_list_one_item_list_zope_ish(self):
        response = self.api_session.get(
            "/@vocabularies/plone.restapi.tests.test_vocabulary?tokens:list=token1"
        )

        self.assertEqual(200, response.status_code)
        response = response.json()
        self.assertEqual(
            response,
            {
                "@id": self.portal_url
                + "/@vocabularies/plone.restapi.tests.test_vocabulary?tokens%3Alist=token1",  # noqa
                "items": [
                    {"title": "Title 1", "token": "token1"},
                ],
                "items_total": 1,
            },
        )

    def test_get_vocabulary_filtered_by_token_list_two_item_list_zope_ish(self):
        response = self.api_session.get(
            "/@vocabularies/plone.restapi.tests.test_vocabulary?tokens:list=token1&tokens:list=token2"
        )

        self.assertEqual(200, response.status_code)
        response = response.json()
        self.assertEqual(
            response,
            {
                "@id": self.portal_url
                + "/@vocabularies/plone.restapi.tests.test_vocabulary?tokens%3Alist=token1&tokens%3Alist=token2",  # noqa
                "items": [
                    {"title": "Title 1", "token": "token1"},
                    {"title": "Title 2", "token": "token2"},
                ],
                "items_total": 2,
            },
        )

    def test_get_vocabulary_filtered_by_token_list_one_item(self):
        response = self.api_session.get(
            "/@vocabularies/plone.restapi.tests.test_vocabulary?tokens=token1"
        )

        self.assertEqual(200, response.status_code)
        response = response.json()
        self.assertEqual(
            response,
            {
                "@id": self.portal_url
                + "/@vocabularies/plone.restapi.tests.test_vocabulary?tokens=token1",  # noqa
                "items": [
                    {"title": "Title 1", "token": "token1"},
                ],
                "items_total": 1,
            },
        )

    def test_get_vocabulary_filtered_by_token_list_two_items(self):
        response = self.api_session.get(
            "/@vocabularies/plone.restapi.tests.test_vocabulary?tokens=token1&tokens=token2"
        )

        self.assertEqual(200, response.status_code)
        response = response.json()
        self.assertEqual(
            response,
            {
                "@id": self.portal_url
                + "/@vocabularies/plone.restapi.tests.test_vocabulary?tokens=token1&tokens=token2",  # noqa
                "items": [
                    {"title": "Title 1", "token": "token1"},
                    {"title": "Title 2", "token": "token2"},
                ],
                "items_total": 2,
            },
        )

    def test_get_unknown_vocabulary(self):
        response = self.api_session.get("/@vocabularies/unknown.vocabulary")

        self.assertEqual(404, response.status_code)
        response = response.json()

        self.assertEqual(
            response,
            {
                "error": {
                    "type": "Not Found",
                    "message": "The vocabulary 'unknown.vocabulary' does not exist",
                }
            },
        )

    def test_get_all_vocabularies(self):
        response = self.api_session.get("/@vocabularies")

        self.assertEqual(200, response.status_code)
        response = response.json()
        self.assertTrue(len(response) > 0)
        self.assertTrue("@id" in list(response[0]))
        self.assertTrue("title" in list(response[0]))
        self.assertEqual(
            [
                {
                    "@id": self.portal_url
                    + "/@vocabularies/plone.restapi.tests.test_vocabulary",  # noqa
                    "title": "plone.restapi.tests.test_vocabulary",
                }
            ],
            [
                x
                for x in response
                if x.get("title") == "plone.restapi.tests.test_vocabulary"
            ],
        )

    def test_context_vocabulary(self):
        api.content.create(
            container=self.portal, id="testdoc", type="Document", title="Document 1"
        )
        transaction.commit()

        context_vocab_name = "plone.restapi.tests.test_context_vocabulary"
        provideUtility(
            test_context_vocabulary_factory,
            provides=IVocabularyFactory,
            name=context_vocab_name,
        )

        response = self.api_session.get(f"testdoc/@vocabularies/{context_vocab_name}")

        gsm = getGlobalSiteManager()
        gsm.unregisterUtility(provided=IVocabularyFactory, name=context_vocab_name)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {
                "@id": self.portal_url
                + "/testdoc/@vocabularies/plone.restapi.tests.test_context_vocabulary",  # noqa
                "items": [
                    {"title": "testdoc", "token": "id"},
                    {"title": "Document 1", "token": "title"},
                ],
                "items_total": 2,
            },
        )

    def test_big_vocabulary_not_batched(self):
        provideUtility(
            test_big_vocabulary_factory,
            provides=IVocabularyFactory,
            name="plone.restapi.tests.test_big_vocabulary",
        )
        response = self.api_session.get(
            "/@vocabularies/plone.restapi.tests.test_big_vocabulary?b_size=-1"
        )

        self.assertEqual(200, response.status_code)
        response = response.json()
        self.assertEqual(len(response["items"]), 100)

    def tearDown(self):
        self.api_session.close()
        gsm = getGlobalSiteManager()
        gsm.unregisterUtility(
            provided=IVocabularyFactory, name="plone.restapi.tests.test_vocabulary"
        )



from datetime import timedelta
from plone.app.contenttypes.interfaces import ILink
from plone.app.textfield.interfaces import IRichText
from plone.app.textfield.value import RichTextValue
from plone.dexterity.interfaces import IDexterityContent
from plone.namedfile.interfaces import INamedField
from plone.restapi.deserializer.utils import path2uid
from plone.restapi.interfaces import IFieldDeserializer
from plone.restapi.services.content.tus import TUSUpload
from pytz import timezone
from pytz import utc
from z3c.form.interfaces import IDataManager
from zope.component import adapter
from zope.component import getMultiAdapter
from zope.component import queryMultiAdapter
from zope.interface import implementer
from zope.publisher.interfaces.browser import IBrowserRequest
from zope.schema.interfaces import IChoice
from zope.schema.interfaces import ICollection
from zope.schema.interfaces import IDatetime
from zope.schema.interfaces import IDict
from zope.schema.interfaces import IField
from zope.schema.interfaces import IFromUnicode
from zope.schema.interfaces import ITextLine
from zope.schema.interfaces import ITime
from zope.schema.interfaces import ITimedelta
from zope.schema.interfaces import IVocabularyTokenized

import codecs
import dateutil
import html as html_parser


@implementer(IFieldDeserializer)
@adapter(IField, IDexterityContent, IBrowserRequest)
class DefaultFieldDeserializer:
    def __init__(self, field, context, request):
        self.field = field
        if IField.providedBy(self.field):
            self.field = self.field.bind(context)
        self.context = context
        self.request = request

    def __call__(self, value):
        if not isinstance(value, str):
            self.field.validate(value)
            return value

        value = IFromUnicode(self.field).fromUnicode(value)
        # IFromUnicode.fromUnicode() will validate, no need to do it twice
        return value


@implementer(IFieldDeserializer)
@adapter(ITextLine, IDexterityContent, IBrowserRequest)
class TextLineFieldDeserializer(DefaultFieldDeserializer):
    def __call__(self, value):
        if isinstance(value, str):
            value = IFromUnicode(self.field).fromUnicode(value)

        # Mimic what z3c.form does in it's BaseDataConverter.
        if isinstance(value, str):
            value = value.strip()
            if value == "":
                value = self.field.missing_value

        self.field.validate(value)
        return value


@implementer(IFieldDeserializer)
@adapter(ITextLine, ILink, IBrowserRequest)
class LinkTextLineFieldDeserializer(TextLineFieldDeserializer):
    def __call__(self, value):
        value = super().__call__(value)
        if self.field.getName() == "remoteUrl":
            portal = getMultiAdapter(
                (self.context, self.context.REQUEST), name="plone_portal_state"
            ).portal()
            portal_url = portal.portal_url()
            if value.startswith(portal_url) or value.startswith("/"):
                value = path2uid(context=self.context, link=value)
        return value


@implementer(IFieldDeserializer)
@adapter(IDatetime, IDexterityContent, IBrowserRequest)
class DatetimeFieldDeserializer(DefaultFieldDeserializer):
    def __call__(self, value):
        # Datetime fields may contain timezone naive or timezone aware
        # objects. Unfortunately the zope.schema.Datetime field does not
        # contain any information if the field value should be timezone naive
        # or timezone aware. While some fields (start, end) store timezone
        # aware objects others (effective, expires) store timezone naive
        # objects.
        # We try to guess the correct deserialization from the current field
        # value.
        dm = queryMultiAdapter((self.context, self.field), IDataManager)
        current = dm.get()
        if current is not None:
            tzinfo = current.tzinfo
        else:
            tzinfo = None

        # This happens when a 'null' is posted for a non-required field.
        if value is None:
            self.field.validate(value)
            return

        # Parse ISO 8601 string with dateutil
        try:
            dt = dateutil.parser.parse(value)
        except ValueError:
            raise ValueError(f"Invalid date: {value}")

        # Convert to TZ aware in UTC
        if dt.tzinfo is not None:
            dt = dt.astimezone(utc)
        else:
            dt = utc.localize(dt)

        # Convert to local TZ aware or naive UTC
        if tzinfo is not None:
            tz = timezone(tzinfo.zone)
            value = tz.normalize(dt.astimezone(tz))
        else:
            value = utc.normalize(dt.astimezone(utc)).replace(tzinfo=None)

        self.field.validate(value)
        return value


@implementer(IFieldDeserializer)
@adapter(IChoice, IDexterityContent, IBrowserRequest)
class ChoiceFieldDeserializer(DefaultFieldDeserializer):
    def __call__(self, value):
        if isinstance(value, dict) and "token" in value:
            value = value["token"]
        if IVocabularyTokenized.providedBy(self.field.vocabulary):
            try:
                value = self.field.vocabulary.getTermByToken(value).value
            except LookupError:
                pass

        self.field.validate(value)
        return value


@implementer(IFieldDeserializer)
@adapter(ICollection, IDexterityContent, IBrowserRequest)
class CollectionFieldDeserializer(DefaultFieldDeserializer):
    def __call__(self, value):
        if not isinstance(value, list):
            value = [value]

        if IField.providedBy(self.field.value_type):
            deserializer = getMultiAdapter(
                (self.field.value_type, self.context, self.request), IFieldDeserializer
            )

            for i, v in enumerate(value):
                if isinstance(v, dict) and "token" in v:
                    v = v["token"]
                value[i] = deserializer(v)

        value = self.field._type(value)
        self.field.validate(value)

        return value


@implementer(IFieldDeserializer)
@adapter(IDict, IDexterityContent, IBrowserRequest)
class DictFieldDeserializer(DefaultFieldDeserializer):
    def __call__(self, value):
        if IField.providedBy(self.field.key_type):
            kdeserializer = getMultiAdapter(
                (self.field.key_type, self.context, self.request), IFieldDeserializer
            )
        else:

            def kdeserializer(k):
                return k

        if IField.providedBy(self.field.value_type):
            vdeserializer = getMultiAdapter(
                (self.field.value_type, self.context, self.request), IFieldDeserializer
            )
        else:

            def vdeserializer(v):
                return v

        new_value = {}
        for k, v in value.items():
            new_value[kdeserializer(k)] = vdeserializer(v)

        self.field.validate(new_value)
        return new_value


@implementer(IFieldDeserializer)
@adapter(ITime, IDexterityContent, IBrowserRequest)
class TimeFieldDeserializer(DefaultFieldDeserializer):
    def __call__(self, value):
        try:
            # Create an ISO 8601 datetime string and parse it with Zope's
            # DateTime module and then convert it to a timezone naive time
            # in local time
            # TODO: should really a timezone naive time be returned?
            # using ``timetz()`` would be timezone aware.
            value = dateutil.parser.parse(value).time()
        except ValueError:
            raise ValueError(f"Invalid time: {value}")

        self.field.validate(value)
        return value


@implementer(IFieldDeserializer)
@adapter(ITimedelta, IDexterityContent, IBrowserRequest)
class TimedeltaFieldDeserializer(DefaultFieldDeserializer):
    def __call__(self, value):
        try:
            value = timedelta(seconds=value)
        except TypeError as e:
            raise ValueError(str(e))

        self.field.validate(value)
        return value


@implementer(IFieldDeserializer)
@adapter(INamedField, IDexterityContent, IBrowserRequest)
class NamedFieldDeserializer(DefaultFieldDeserializer):
    def __call__(self, value):
        content_type = "application/octet-stream"
        filename = None
        if isinstance(value, dict):
            if "data" not in value:
                # We are probably pushing the contents of a previous GET
                # That contain the read representation of the file
                # with the 'download' key so we return the same stored file
                return getattr(self.field.context, self.field.__name__)

            content_type = value.get("content-type", content_type)
            filename = value.get("filename", filename)
            data = value.get("data", "")
            if isinstance(data, str):
                data = data.encode("utf-8")
            if "encoding" in value:
                data = codecs.decode(data, value["encoding"])
            if isinstance(data, str):
                data = data.encode("utf-8")
        elif isinstance(value, TUSUpload):
            content_type = value.metadata().get("content-type", content_type)
            filename = value.metadata().get("filename", filename)
            data = value.open()
        else:
            data = value

        # Convert if we have data
        if data:
            value = self.field._type(
                data=data, contentType=content_type, filename=filename
            )
        else:
            value = None

        # Always validate to check for required fields
        self.field.validate(value)
        return value


@implementer(IFieldDeserializer)
@adapter(IRichText, IDexterityContent, IBrowserRequest)
class RichTextFieldDeserializer(DefaultFieldDeserializer):
    def __call__(self, value):
        content_type = self.field.default_mime_type
        encoding = "utf8"
        if isinstance(value, dict):
            content_type = value.get("content-type", content_type)
            encoding = value.get("encoding", encoding)
            data = value.get("data", "")
        elif isinstance(value, TUSUpload):
            content_type = value.metadata().get("content-type", content_type)
            with open(value.filepath, "rb") as f:
                data = f.read().decode("utf8")
        else:
            data = value
        value = RichTextValue(
            raw=html_parser.unescape(data),
            mimeType=content_type,
            outputMimeType=self.field.output_mime_type,
            encoding=encoding,
        )
        self.field.validate(value)
        return value



from collections import deque
from copy import deepcopy
from plone import api
from plone.restapi.behaviors import IBlocks
from plone.restapi.deserializer.dxfields import DefaultFieldDeserializer
from plone.restapi.deserializer.utils import path2uid
from plone.restapi.interfaces import IBlockFieldDeserializationTransformer
from plone.restapi.interfaces import IFieldDeserializer
from plone.schema import IJSONField
from Products.CMFPlone.interfaces import IPloneSiteRoot
from zope.component import adapter
from zope.component import subscribers
from zope.interface import implementer
from zope.publisher.interfaces.browser import IBrowserRequest

import os


def iterate_children(value):
    """iterate_children.

    :param value:
    """
    queue = deque(value)
    while queue:
        child = queue.pop()
        yield child
        if child.get("children"):
            queue.extend(child["children"] or [])


@implementer(IFieldDeserializer)
@adapter(IJSONField, IBlocks, IBrowserRequest)
class BlocksJSONFieldDeserializer(DefaultFieldDeserializer):
    def _transform(self, blocks):
        for id, block_value in blocks.items():
            self.handle_subblocks(block_value)
            block_type = block_value.get("@type", "")
            handlers = []
            for h in subscribers(
                (self.context, self.request), IBlockFieldDeserializationTransformer
            ):
                if h.block_type == block_type or h.block_type is None:
                    h.blockid = id
                    handlers.append(h)

            for handler in sorted(handlers, key=lambda h: h.order):
                block_value = handler(block_value)

            blocks[id] = block_value

        return blocks

    def handle_subblocks(self, block_value):
        if "data" in block_value:
            if isinstance(block_value["data"], dict):
                if "blocks" in block_value["data"]:
                    block_value["data"]["blocks"] = self._transform(
                        block_value["data"]["blocks"]
                    )

        if "blocks" in block_value:
            block_value["blocks"] = self._transform(block_value["blocks"])

    def __call__(self, value):
        value = super().__call__(value)

        if self.field.getName() == "blocks":
            for id, block_value in value.items():
                self.handle_subblocks(block_value)
                block_type = block_value.get("@type", "")

                handlers = []
                for h in subscribers(
                    (self.context, self.request),
                    IBlockFieldDeserializationTransformer,
                ):
                    if h.block_type == block_type or h.block_type is None:
                        h.blockid = id
                        handlers.append(h)

                for handler in sorted(handlers, key=lambda h: h.order):
                    if not getattr(handler, "disabled", False):
                        block_value = handler(block_value)

                value[id] = block_value

        return value


class ResolveUIDDeserializerBase:
    """The "url" smart block field.

    This is a generic handler. In all blocks, it converts any "url"
    field from using resolveuid to an "absolute" URL
    """

    order = 1
    block_type = None
    fields = ["url", "href"]
    disabled = os.environ.get("disable_transform_resolveuid", False)

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def __call__(self, block):
        # Convert absolute links to resolveuid
        for field in self.fields:
            link = block.get(field, "")
            if link and isinstance(link, str):
                block[field] = path2uid(context=self.context, link=link)
            elif link and isinstance(link, list):
                # Detect if it has an object inside with an "@id" key (object_widget)
                if len(link) > 0 and isinstance(link[0], dict) and "@id" in link[0]:
                    result = []
                    for item in link:
                        item_clone = deepcopy(item)
                        item_clone["@id"] = path2uid(
                            context=self.context, link=item_clone["@id"]
                        )
                        result.append(item_clone)

                    block[field] = result
                elif len(link) > 0 and isinstance(link[0], str):
                    block[field] = [
                        path2uid(context=self.context, link=item) for item in link
                    ]
        return block


class TextBlockDeserializerBase:
    order = 100
    block_type = "text"
    disabled = os.environ.get("disable_transform_resolveuid", False)

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def __call__(self, block):
        # Convert absolute links to resolveuid
        #   http://localhost:55001/plone/link-target
        #   ->
        #   ../resolveuid/023c61b44e194652804d05a15dc126f4
        entity_map = block.get("text", {}).get("entityMap", {})
        for entity in entity_map.values():
            if entity.get("type") == "LINK":
                href = entity.get("data", {}).get("url", "")
                entity["data"]["url"] = path2uid(context=self.context, link=href)
        return block


class HTMLBlockDeserializerBase:
    order = 100
    block_type = "html"
    disabled = os.environ.get("disable_transform_html", False)

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def __call__(self, block):

        portal_transforms = api.portal.get_tool(name="portal_transforms")
        raw_html = block.get("html", "")
        data = portal_transforms.convertTo(
            "text/x-html-safe", raw_html, mimetype="text/html"
        )
        html = data.getData()
        block["html"] = html

        return block


class ImageBlockDeserializerBase:
    order = 100
    block_type = "image"
    disabled = os.environ.get("disable_transform_resolveuid", False)

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def __call__(self, block):
        url = block.get("url", "")
        block["url"] = path2uid(context=self.context, link=url)
        return block


@adapter(IBlocks, IBrowserRequest)
@implementer(IBlockFieldDeserializationTransformer)
class ResolveUIDDeserializer(ResolveUIDDeserializerBase):
    """Deserializer for content-types that implements IBlocks behavior"""


@adapter(IPloneSiteRoot, IBrowserRequest)
@implementer(IBlockFieldDeserializationTransformer)
class ResolveUIDDeserializerRoot(ResolveUIDDeserializerBase):
    """Deserializer for site root"""


@adapter(IBlocks, IBrowserRequest)
@implementer(IBlockFieldDeserializationTransformer)
class TextBlockDeserializer(TextBlockDeserializerBase):
    """Deserializer for content-types that implements IBlocks behavior"""


@adapter(IPloneSiteRoot, IBrowserRequest)
@implementer(IBlockFieldDeserializationTransformer)
class TextBlockDeserializerRoot(TextBlockDeserializerBase):
    """Deserializer for site root"""


@adapter(IBlocks, IBrowserRequest)
@implementer(IBlockFieldDeserializationTransformer)
class HTMLBlockDeserializer(HTMLBlockDeserializerBase):
    """Deserializer for content-types that implements IBlocks behavior"""


@adapter(IPloneSiteRoot, IBrowserRequest)
@implementer(IBlockFieldDeserializationTransformer)
class HTMLBlockDeserializerRoot(HTMLBlockDeserializerBase):
    """Deserializer for site root"""


@adapter(IBlocks, IBrowserRequest)
@implementer(IBlockFieldDeserializationTransformer)
class ImageBlockDeserializer(ImageBlockDeserializerBase):
    """Deserializer for content-types that implements IBlocks behavior"""


@adapter(IPloneSiteRoot, IBrowserRequest)
@implementer(IBlockFieldDeserializationTransformer)
class ImageBlockDeserializerRoot(ImageBlockDeserializerBase):
    """Deserializer for site root"""


def transform_links(context, value, transformer):
    """Convert absolute links to resolveuid
    http://localhost:55001/plone/link-target
    ->
    ../resolveuid/023c61b44e194652804d05a15dc126f4"""
    data = value.get("data", {})
    if data.get("link", {}).get("internal", {}).get("internal_link"):
        internal_link = data["link"]["internal"]["internal_link"]
        for link in internal_link:
            link["@id"] = transformer(context, link["@id"])


class SlateBlockTransformer:
    """SlateBlockTransformer."""

    field = "value"

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def __call__(self, block):
        value = (block or {}).get(self.field, [])
        children = iterate_children(value or [])

        for child in children:
            node_type = child.get("type")
            if node_type:
                handler = getattr(self, f"handle_{node_type}", None)
                if handler:
                    handler(child)

        return block


class SlateBlockDeserializerBase(SlateBlockTransformer):
    """SlateBlockDeserializerBase."""

    order = 100
    block_type = "slate"
    disabled = os.environ.get("disable_transform_resolveuid", False)

    def handle_a(self, child):
        transform_links(self.context, child, transformer=path2uid)

    def handle_link(self, child):
        if child.get("data", {}).get("url"):
            child["data"]["url"] = path2uid(self.context, child["data"]["url"])


@adapter(IBlocks, IBrowserRequest)
@implementer(IBlockFieldDeserializationTransformer)
class SlateBlockDeserializer(SlateBlockDeserializerBase):
    """Deserializer for content-types that implements IBlocks behavior"""


@adapter(IPloneSiteRoot, IBrowserRequest)
@implementer(IBlockFieldDeserializationTransformer)
class SlateBlockDeserializerRoot(SlateBlockDeserializerBase):
    """Deserializer for site root"""



from AccessControl import getSecurityManager
from Acquisition import aq_inner
from io import BytesIO
from OFS.Image import Image
from plone.restapi.services import Service
from Products.CMFCore.permissions import SetOwnPassword
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.interfaces import ISecuritySchema
from Products.CMFPlone.utils import set_own_login_name
from Products.PlonePAS.tools.membership import default_portrait
from Products.PlonePAS.utils import scale_image
from zope.component import getAdapter
from zope.component.hooks import getSite
from zope.interface import alsoProvides
from zope.interface import implementer
from zope.publisher.interfaces import IPublishTraverse

import codecs
import json
import plone


@implementer(IPublishTraverse)
class UsersPatch(Service):
    """Updates an existing user."""

    def __init__(self, context, request):
        super().__init__(context, request)
        self.params = []

    def publishTraverse(self, request, name):
        # Consume any path segments after /@users as parameters
        self.params.append(name)
        return self

    @property
    def _get_user_id(self):
        if len(self.params) != 1:
            raise Exception("Must supply exactly one parameter (user id)")
        return self.params[0]

    def _get_user(self, user_id):
        portal = getSite()
        portal_membership = getToolByName(portal, "portal_membership")
        return portal_membership.getMemberById(user_id)

    def _change_user_password(self, user, value):
        acl_users = getToolByName(self.context, "acl_users")
        acl_users.userSetPassword(user.getUserId(), value)

    def reply(self):
        user_settings_to_update = json.loads(self.request.get("BODY", "{}"))
        user = self._get_user(self._get_user_id)

        # Disable CSRF protection
        if "IDisableCSRFProtection" in dir(plone.protect.interfaces):
            alsoProvides(self.request, plone.protect.interfaces.IDisableCSRFProtection)

        security = getAdapter(self.context, ISecuritySchema)

        if self.can_manage_users:
            for key, value in user_settings_to_update.items():
                if key == "password":
                    self._change_user_password(user, value)
                elif key == "username":
                    set_own_login_name(user, value)
                else:
                    # If the portrait is already set but has not been changed change it,
                    # then the serialized value comes again in the request as a string,
                    # no data on it, then we should not set it since it will fail
                    if key == "portrait" and isinstance(value, dict):
                        self.set_member_portrait(user, value)
                    user.setMemberProperties(mapping={key: value}, force_empty=True)

            roles = user_settings_to_update.get("roles", {})
            if roles:
                to_add = [key for key, enabled in roles.items() if enabled]
                to_remove = [key for key, enabled in roles.items() if not enabled]

                target_roles = set(user.getRoles()) - set(to_remove)
                target_roles = target_roles | set(to_add)

                acl_users = getToolByName(self.context, "acl_users")
                acl_users.userFolderEditUser(
                    principal_id=user.id,
                    password=None,
                    roles=target_roles,
                    domains=user.getDomains(),
                )
        elif self._get_current_user == self._get_user_id:
            for key, value in user_settings_to_update.items():
                if (
                    key == "password"
                    and security.enable_user_pwd_choice
                    and self.can_set_own_password
                ):
                    self._change_user_password(user, value)
                else:
                    # If the portrait is already set but has not been changed change it,
                    # then the serialized value comes again in the request as a string,
                    # no data on it, then we should not set it since it will fail
                    if key == "portrait" and isinstance(value, dict):
                        self.set_member_portrait(user, value)
                    user.setMemberProperties(mapping={key: value}, force_empty=True)

        else:
            if self._is_anonymous:
                return self._error(
                    401,
                    "Unauthorized",
                    "You are not authorized to perform this " "action",
                )
            else:
                return self._error(
                    403, "Forbidden", "You can't update the " "properties of this user"
                )

        return self.reply_no_content()

    @property
    def can_manage_users(self):
        sm = getSecurityManager()
        return sm.checkPermission("plone.app.controlpanel.UsersAndGroups", self.context)

    @property
    def can_set_own_password(self):
        sm = getSecurityManager()
        return sm.checkPermission(SetOwnPassword, self.context)

    def _error(self, status, type, message):
        self.request.response.setStatus(status)
        return {"error": {"type": type, "message": message}}

    @property
    def _get_current_user(self):
        portal = getSite()
        portal_membership = getToolByName(portal, "portal_membership")
        return portal_membership.getAuthenticatedMember().getId()

    @property
    def _is_anonymous(self):
        portal = getSite()
        portal_membership = getToolByName(portal, "portal_membership")
        return portal_membership.isAnonymousUser()

    def set_member_portrait(self, user, portrait):
        portal = getSite()
        portal_membership = getToolByName(portal, "portal_membership")
        safe_id = portal_membership._getSafeMemberId(user.getId())

        if portrait is None:
            previous = portal_membership.getPersonalPortrait(safe_id)
            default_portrait_value = getattr(portal, default_portrait, None)
            if aq_inner(previous) != aq_inner(default_portrait_value):
                portal_membership.deletePersonalPortrait(str(safe_id))
            return

        content_type = "application/octet-stream"
        filename = None

        content_type = portrait.get("content-type", content_type)
        filename = portrait.get("filename", filename)
        data = portrait.get("data")
        if isinstance(data, str):
            data = data.encode("utf-8")
        if "encoding" in portrait:
            data = codecs.decode(data, portrait["encoding"])

        if portrait.get("scale", False):
            # Only scale if the scale (default Plone behavior) boolean is set
            # This should be handled by the core in the future
            scaled, mimetype = scale_image(BytesIO(data))
        else:
            # Normally, the scale and cropping is going to be handled in the
            # frontend
            scaled = data

        portrait = Image(id=safe_id, file=scaled, title="")
        membertool = getToolByName(self, "portal_memberdata")
        membertool._setPortrait(portrait, safe_id)



from plone.restapi.deserializer import json_body
from plone.restapi.services import Service
from Products.CMFCore.utils import getToolByName
from zExceptions import BadRequest
from zope.component.hooks import getSite
from zope.interface import alsoProvides
from zope.interface import implementer
from zope.publisher.interfaces import IPublishTraverse

import plone


@implementer(IPublishTraverse)
class GroupsPatch(Service):
    """Updates an existing group."""

    def __init__(self, context, request):
        super().__init__(context, request)
        self.params = []

    def publishTraverse(self, request, name):
        # Consume any path segments after /@groups as parameters
        self.params.append(name)
        return self

    @property
    def _get_group_id(self):
        if len(self.params) != 1:
            raise Exception("Must supply exactly one parameter (group id)")
        return self.params[0]

    def _get_group(self, group_id):
        portal = getSite()
        portal_groups = getToolByName(portal, "portal_groups")
        return portal_groups.getGroupById(group_id)

    def reply(self):
        data = json_body(self.request)
        group = self._get_group(self._get_group_id)

        if not group:
            raise BadRequest("Trying to update a non-existing group.")

        title = data.get("title", None)
        description = data.get("description", None)
        roles = data.get("roles", None)
        groups = data.get("groups", None)
        users = data.get("users", {})

        # Disable CSRF protection
        if "IDisableCSRFProtection" in dir(plone.protect.interfaces):
            alsoProvides(self.request, plone.protect.interfaces.IDisableCSRFProtection)

        portal_groups = getToolByName(self.context, "portal_groups")

        portal_groups.editGroup(
            self._get_group_id,
            roles=roles,
            groups=groups,
            title=title,
            description=description,
        )

        properties = {}
        for id, property in group.propertyItems():
            if data.get(id, False):
                properties[id] = data[id]

        group.setGroupProperties(properties)

        # Add/remove members
        memberids = group.getGroupMemberIds()
        for userid, allow in users.items():
            if allow:
                if userid not in memberids:
                    group.addMember(userid)
            else:
                if userid in memberids:
                    group.removeMember(userid)

        return self.reply_no_content()



from plone.registry.interfaces import IRegistry
from plone.restapi.batching import HypermediaBatch
from plone.restapi.interfaces import ISerializeToJson
from plone.restapi.types.interfaces import IJsonSchemaProvider
from zope.component import adapter
from zope.component import getMultiAdapter
from zope.interface import implementer
from zope.publisher.interfaces import IRequest


@implementer(ISerializeToJson)
@adapter(IRegistry, IRequest)
class SerializeRegistryToJson:
    def __init__(self, registry, request):
        self.registry = registry
        self.request = request

    def __call__(self):
        records = self.registry.records
        # Batch keys, because that is a simple BTree
        batch = HypermediaBatch(self.request, list(records))

        results = {}
        results["@id"] = batch.canonical_url
        results["items_total"] = batch.items_total
        if batch.links:
            results["batching"] = batch.links

        def make_item(key):
            record = records[key]
            schema = getMultiAdapter(
                (record.field, record, self.request), IJsonSchemaProvider
            )
            data = {"name": key, "value": self.registry[key]}
            __traceback_info__ = (record, record.field, schema)
            data["schema"] = {"properties": schema.get_schema()}
            return data

        results["items"] = [make_item(key) for key in batch]
        return results



from plone.restapi.services import Service
from Products.CMFCore.utils import getToolByName
from Products.PluggableAuthService.interfaces.plugins import (
    IAuthenticationPlugin,
)  # noqa
from zope.interface import alsoProvides

import plone.protect.interfaces


class Renew(Service):
    """Renew authentication token"""

    def reply(self):
        plugin = None
        acl_users = getToolByName(self, "acl_users")
        plugins = acl_users._getOb("plugins")
        authenticators = plugins.listPlugins(IAuthenticationPlugin)
        for id_, authenticator in authenticators:
            if authenticator.meta_type == "JWT Authentication Plugin":
                plugin = authenticator
                break

        if plugin is None:
            self.request.response.setStatus(501)
            return dict(
                error=dict(
                    type="Renew failed",
                    message="JWT authentication plugin not installed.",
                )
            )

        # Disable CSRF protection
        if "IDisableCSRFProtection" in dir(plone.protect.interfaces):
            alsoProvides(self.request, plone.protect.interfaces.IDisableCSRFProtection)

        mtool = getToolByName(self.context, "portal_membership")
        if bool(mtool.isAnonymousUser()):
            # Don't generate authentication tokens for anonymous users.
            self.request.response.setStatus(401)
            return dict(
                error=dict(
                    type="Invalid or expired authentication token",
                    message="The authentication token is invalid or expired.",
                )
            )
        user = mtool.getAuthenticatedMember()
        payload = {}
        payload["fullname"] = user.getProperty("fullname")
        new_token = plugin.create_token(user.getId(), data=payload)
        if plugin.store_tokens and self.request._auth:
            old_token = self.request._auth[7:]
            plugin.delete_token(old_token)
        return {"token": new_token}



from plone import api
from plone.app.testing import setRoles
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import SITE_OWNER_PASSWORD
from plone.app.testing import TEST_USER_ID
from plone.restapi.testing import PLONE_RESTAPI_DX_FUNCTIONAL_TESTING
from plone.restapi.testing import RelativeSession

import transaction
import unittest


class TestSourcesEndpoint(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_FUNCTIONAL_TESTING

    maxDiff = None

    def setUp(self):
        self.app = self.layer["app"]
        self.portal = self.layer["portal"]
        self.portal_url = self.portal.absolute_url()
        setRoles(self.portal, TEST_USER_ID, ["Manager"])

        self.api_session = RelativeSession(self.portal_url, test=self)
        self.api_session.headers.update({"Accept": "application/json"})
        self.api_session.auth = (SITE_OWNER_NAME, SITE_OWNER_PASSWORD)

        self.doc = api.content.create(
            container=self.portal,
            id="testdoc",
            type="DXTestDocument",
            title="Document 1",
        )
        transaction.commit()

    def test_get_source(self):
        response = self.api_session.get(
            "%s/@sources/test_choice_with_source" % self.doc.absolute_url()
        )

        self.assertEqual(200, response.status_code)
        response = response.json()
        self.assertEqual(
            response,
            {
                "@id": self.doc.absolute_url()
                + "/@sources/test_choice_with_source",  # noqa
                "items": [
                    {"title": "Title 1", "token": "token1"},
                    {"title": "Title 2", "token": "token2"},
                    {"title": "Title 3", "token": "token3"},
                ],
                "items_total": 3,
            },
        )

    def test_get_source_batched(self):
        response = self.api_session.get(
            "%s/@sources/test_choice_with_source?b_size=1" % self.doc.absolute_url()
        )

        self.assertEqual(200, response.status_code)
        response = response.json()
        self.assertEqual(
            response,
            {
                "@id": self.doc.absolute_url()
                + "/@sources/test_choice_with_source",  # noqa
                "batching": {
                    "@id": self.doc.absolute_url()
                    + "/@sources/test_choice_with_source?b_size=1",  # noqa
                    "first": self.doc.absolute_url()
                    + "/@sources/test_choice_with_source?b_start=0&b_size=1",  # noqa
                    "last": self.doc.absolute_url()
                    + "/@sources/test_choice_with_source?b_start=2&b_size=1",  # noqa
                    "next": self.doc.absolute_url()
                    + "/@sources/test_choice_with_source?b_start=1&b_size=1",  # noqa
                },
                "items": [{"title": "Title 1", "token": "token1"}],
                "items_total": 3,
            },
        )

    def test_get_source_filtered_by_title(self):
        response = self.api_session.get(
            "%s/@sources/test_choice_with_source?title=2" % self.doc.absolute_url()
        )

        self.assertEqual(200, response.status_code)
        response = response.json()
        self.assertEqual(
            response,
            {
                "@id": self.doc.absolute_url()
                + "/@sources/test_choice_with_source?title=2",  # noqa
                "items": [{"title": "Title 2", "token": "token2"}],
                "items_total": 1,
            },
        )

    def test_get_source_filtered_by_token(self):
        response = self.api_session.get(
            "%s/@sources/test_choice_with_source?token=token1" % self.doc.absolute_url()
        )

        self.assertEqual(200, response.status_code)
        response = response.json()
        self.assertEqual(
            response,
            {
                "@id": self.doc.absolute_url()
                + "/@sources/test_choice_with_source?token=token1",  # noqa
                "items": [{"title": "Title 1", "token": "token1"}],
                "items_total": 1,
            },
        )

    def test_get_source_filtered_by_token_partial_not_match(self):
        response = self.api_session.get(
            "%s/@sources/test_choice_with_source?token=token" % self.doc.absolute_url()
        )

        self.assertEqual(200, response.status_code)
        response = response.json()
        self.assertEqual(
            response,
            {
                "@id": self.doc.absolute_url()
                + "/@sources/test_choice_with_source?token=token",  # noqa
                "items": [],
                "items_total": 0,
            },
        )

    def test_get_source_filtered_by_title_and_token_returns_error(self):
        response = self.api_session.get(
            "%s/@sources/test_choice_with_source?token=token1&title=Title"
            % self.doc.absolute_url()  # noqa
        )

        self.assertEqual(400, response.status_code)
        response = response.json()
        self.assertEqual(
            response.get("error"),
            {
                "type": "Invalid parameters",
                "message": "You can not filter by title and token at the same time.",  # noqa
            },
        )

    def test_get_non_iterable_source_returns_error(self):
        response = self.api_session.get(
            "%s/@sources/test_choice_with_non_iterable_source" % self.doc.absolute_url()
        )

        self.assertEqual(400, response.status_code)
        response = response.json()
        self.assertEqual(
            response.get("error"),
            {
                "type": "Bad Request",
                "message": "Source for field 'test_choice_with_non_iterable_source' is not iterable. ",
            },
        )

    def test_get_source_for_unknown_field(self):
        response = self.api_session.get(
            "%s/@sources/unknown_field" % self.doc.absolute_url()
        )

        self.assertEqual(404, response.status_code)
        response = response.json()

        self.assertEqual(
            response.get("error"),
            {"type": "Not Found", "message": "No such field: 'unknown_field'"},
        )

    def test_context_source(self):
        self.doc.title = "Foo Bar Baz"
        transaction.commit()

        response = self.api_session.get(
            "%s/@sources/test_choice_with_context_source" % self.doc.absolute_url()
        )

        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {
                "@id": self.portal_url
                + "/testdoc/@sources/test_choice_with_context_source",  # noqa
                "items": [
                    {"token": "foo", "title": "Foo"},
                    {"token": "bar", "title": "Bar"},
                    {"token": "baz", "title": "Baz"},
                ],
                "items_total": 3,
            },
        )

    def test_source_filtered_by_non_ascii_title(self):
        self.doc.title = "Bär"
        transaction.commit()

        response = self.api_session.get(
            "%s/@sources/test_choice_with_context_source?title=b%%C3%%A4r"
            % self.doc.absolute_url()
        )

        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {
                "@id": self.portal_url
                + "/testdoc/@sources/test_choice_with_context_source?title=b%C3%A4r",  # noqa
                "items": [{"token": "b=C3=A4r", "title": "B\xe4r"}],
                "items_total": 1,
            },
        )

    def tearDown(self):
        self.api_session.close()



from AccessControl import getSecurityManager
from AccessControl.Permissions import use_mailhost_services
from email.mime.text import MIMEText
from plone.registry.interfaces import IRegistry
from plone.restapi import _
from plone.restapi.deserializer import json_body
from plone.restapi.services import Service
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.interfaces import ISiteSchema
from Products.CMFPlone.interfaces.controlpanel import IMailSchema
from smtplib import SMTPException
from zope.component import getMultiAdapter
from zope.component import getUtility
from zope.interface import alsoProvides

import plone


class EmailSendPost(Service):
    def reply(self):
        data = json_body(self.request)

        send_to_address = data.get("to", None)
        sender_from_address = data.get("from", None)
        message = data.get("message", None)
        sender_fullname = data.get("name", "")
        subject = data.get("subject", "")

        if not send_to_address or not sender_from_address or not message:
            self.request.response.setStatus(400)
            return dict(
                error=dict(
                    type="BadRequest",
                    message='Missing "to", "from" or "message" parameters',
                )
            )

        overview_controlpanel = getMultiAdapter(
            (self.context, self.request), name="overview-controlpanel"
        )
        if overview_controlpanel.mailhost_warning():
            self.request.response.setStatus(400)
            return dict(
                error=dict(type="BadRequest", message="MailHost is not configured.")
            )

        sm = getSecurityManager()
        if not sm.checkPermission(use_mailhost_services, self.context):
            pm = getToolByName(self.context, "portal_membership")
            if bool(pm.isAnonymousUser()):
                self.request.response.setStatus(401)
                error_type = "Unauthorized"
            else:
                self.request.response.setStatus(403)
                error_type = "Forbidden"
            return dict(error=dict(type=error_type, message=message))

        # Disable CSRF protection
        if "IDisableCSRFProtection" in dir(plone.protect.interfaces):
            alsoProvides(self.request, plone.protect.interfaces.IDisableCSRFProtection)

        registry = getUtility(IRegistry)
        mail_settings = registry.forInterface(IMailSchema, prefix="plone")
        from_address = mail_settings.email_from_address
        encoding = registry.get("plone.email_charset", "utf-8")
        host = getToolByName(self.context, "MailHost")
        registry = getUtility(IRegistry)
        site_settings = registry.forInterface(ISiteSchema, prefix="plone", check=False)
        portal_title = site_settings.site_title

        if not subject:
            if not sender_fullname:
                subject = self.context.translate(
                    _(
                        "A portal user via ${portal_title}",
                        mapping={"portal_title": portal_title},
                    )
                )
            else:
                subject = self.context.translate(
                    _(
                        "${sender_fullname} via ${portal_title}",
                        mapping={
                            "sender_fullname": sender_fullname,
                            "portal_title": portal_title,
                        },
                    )
                )

        message_intro = self.context.translate(
            _(
                "You are receiving this mail because ${sender_fullname} sent this message via the site ${portal_title}:",  # noqa
                mapping={
                    "sender_fullname": sender_fullname or "a portal user",
                    "portal_title": portal_title,
                },
            )
        )

        message = f"{message_intro} \n {message}"

        message = MIMEText(message, "plain", encoding)
        message["Reply-To"] = sender_from_address
        try:
            host.send(
                message,
                send_to_address,
                from_address,
                subject=subject,
                charset=encoding,
            )

        except (SMTPException, RuntimeError):
            plone_utils = getToolByName(self.context, "plone_utils")
            exception = plone_utils.exceptionString()
            message = f"Unable to send mail: {exception}"

            self.request.response.setStatus(500)
            return dict(error=dict(type="InternalServerError", message=message))

        return self.reply_no_content()



from OFS.interfaces import IObjectWillBeAddedEvent
from plone.app.testing import login
from plone.app.testing import setRoles
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import SITE_OWNER_PASSWORD
from plone.app.testing import TEST_USER_ID
from plone.app.testing import TEST_USER_NAME
from plone.app.testing import TEST_USER_PASSWORD
from plone.restapi.testing import PLONE_RESTAPI_DX_FUNCTIONAL_TESTING
from Products.CMFCore.utils import getToolByName
from zope.component import getGlobalSiteManager
from zope.lifecycleevent.interfaces import IObjectAddedEvent
from zope.lifecycleevent.interfaces import IObjectCreatedEvent
from zope.lifecycleevent.interfaces import IObjectModifiedEvent

import requests
import transaction
import unittest


class TestFolderCreate(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_FUNCTIONAL_TESTING

    def setUp(self):
        self.app = self.layer["app"]
        self.portal = self.layer["portal"]
        self.portal_url = self.portal.absolute_url()
        setRoles(self.portal, TEST_USER_ID, ["Member"])
        login(self.portal, SITE_OWNER_NAME)
        self.portal.invokeFactory("Folder", id="folder1", title="My Folder")
        wftool = getToolByName(self.portal, "portal_workflow")
        wftool.doActionFor(self.portal.folder1, "publish")
        transaction.commit()

    def test_post_to_folder_creates_document(self):
        response = requests.post(
            self.portal.folder1.absolute_url(),
            headers={"Accept": "application/json"},
            auth=(SITE_OWNER_NAME, SITE_OWNER_PASSWORD),
            json={"@type": "Document", "id": "mydocument", "title": "My Document"},
        )
        self.assertEqual(201, response.status_code)
        transaction.begin()
        self.assertEqual("My Document", self.portal.folder1.mydocument.Title())
        self.assertEqual("Document", response.json().get("@type"))
        self.assertEqual("mydocument", response.json().get("id"))
        self.assertEqual("My Document", response.json().get("title"))

        expected_url = self.portal_url + "/folder1/mydocument"
        self.assertEqual(expected_url, response.json().get("@id"))

    def test_post_to_folder_creates_folder(self):
        response = requests.post(
            self.portal.folder1.absolute_url(),
            headers={"Accept": "application/json"},
            auth=(SITE_OWNER_NAME, SITE_OWNER_PASSWORD),
            json={"@type": "Folder", "id": "myfolder", "title": "My Folder"},
        )
        self.assertEqual(201, response.status_code)
        transaction.begin()
        self.assertEqual("My Folder", self.portal.folder1.myfolder.Title())
        self.assertEqual("Folder", response.json().get("@type"))
        self.assertEqual("myfolder", response.json().get("id"))
        self.assertEqual("My Folder", response.json().get("title"))

        expected_url = self.portal_url + "/folder1/myfolder"
        self.assertEqual(expected_url, response.json().get("@id"))

    def test_post_without_type_returns_validation_error(self):
        response = requests.post(
            self.portal.folder1.absolute_url(),
            headers={"Accept": "application/json"},
            auth=(SITE_OWNER_NAME, SITE_OWNER_PASSWORD),
            json={"id": "mydocument", "title": "My Document"},
        )
        self.assertEqual(400, response.status_code)
        self.assertIn("Property '@type' is required", response.text)

    def test_post_with_invalid_type_returns_validation_error(self):
        response = requests.post(
            self.portal.folder1.absolute_url(),
            headers={"Accept": "application/json"},
            auth=(SITE_OWNER_NAME, SITE_OWNER_PASSWORD),
            json={
                "@type": "NonExistentType",
                "id": "mydocument",
                "title": "My Document",
            },
        )
        self.assertEqual(403, response.status_code)
        self.assertEqual(
            "Invalid '@type' parameter. No content type with the name 'NonExistentType' found",
            response.json().get("error").get("message"),
        )

    def test_post_with_empty_type_returns_validation_error(self):
        response = requests.post(
            self.portal.folder1.absolute_url(),
            headers={"Accept": "application/json"},
            auth=(SITE_OWNER_NAME, SITE_OWNER_PASSWORD),
            json={
                "@type": "",
                "id": "mydocument",
                "title": "My Document",
            },
        )
        self.assertEqual(400, response.status_code)
        self.assertIn("Property '@type' is required", response.text)

    def test_post_without_id_creates_id_from_title(self):
        response = requests.post(
            self.portal.folder1.absolute_url(),
            headers={"Accept": "application/json"},
            auth=(SITE_OWNER_NAME, SITE_OWNER_PASSWORD),
            json={"@type": "Document", "title": "My Document"},
        )
        self.assertEqual(201, response.status_code)
        transaction.begin()
        self.assertIn("my-document", self.portal.folder1)

    def test_post_without_id_creates_id_from_filename(self):
        response = requests.post(
            self.portal.folder1.absolute_url(),
            headers={"Accept": "application/json"},
            auth=(SITE_OWNER_NAME, SITE_OWNER_PASSWORD),
            json={
                "@type": "File",
                "title": "My File",
                "file": {
                    "filename": "test.txt",
                    "data": "Spam and Eggs",
                    "content_type": "text/plain",
                },
            },
        )
        self.assertEqual(201, response.status_code)
        transaction.begin()
        self.assertIn("test.txt", self.portal.folder1)

    def test_post_with_id_already_in_use_returns_400(self):
        self.portal.folder1.invokeFactory("Document", "mydocument")
        transaction.commit()
        response = requests.post(
            self.portal.folder1.absolute_url(),
            headers={"Accept": "application/json"},
            auth=(SITE_OWNER_NAME, SITE_OWNER_PASSWORD),
            json={"@type": "Document", "id": "mydocument", "title": "My Document"},
        )
        self.assertEqual(400, response.status_code)

    def test_post_to_folder_returns_401_unauthorized(self):
        response = requests.post(
            self.portal.folder1.absolute_url(),
            headers={"Accept": "application/json"},
            auth=(TEST_USER_NAME, TEST_USER_PASSWORD),
            json={"@type": "Document", "id": "mydocument", "title": "My Document"},
        )
        self.assertEqual(401, response.status_code)

    def test_post_to_folder_without_add_permission_returns_403_forbidden(self):
        self.portal.folder1.manage_permission(
            "plone.app.contenttypes: Add Document", [], acquire=False
        )
        transaction.commit()
        response = requests.post(
            self.portal.folder1.absolute_url(),
            headers={"Accept": "application/json"},
            auth=(SITE_OWNER_NAME, SITE_OWNER_PASSWORD),
            json={"@type": "Document", "id": "mydocument", "title": "My Document"},
        )
        self.assertEqual(403, response.status_code)

    def test_post_to_folder_fires_proper_events(self):
        sm = getGlobalSiteManager()
        fired_events = []

        def record_event(event):
            fired_events.append(event.__class__.__name__)

        sm.registerHandler(record_event, (IObjectCreatedEvent,))
        sm.registerHandler(record_event, (IObjectWillBeAddedEvent,))
        sm.registerHandler(record_event, (IObjectAddedEvent,))
        sm.registerHandler(record_event, (IObjectModifiedEvent,))

        requests.post(
            self.portal.folder1.absolute_url(),
            headers={"Accept": "application/json"},
            auth=(SITE_OWNER_NAME, SITE_OWNER_PASSWORD),
            json={
                "@type": "Document",
                "id": "mydocument",
                "title": "My Document",
                "description": "123",
            },
        )

        self.assertEqual(
            fired_events,
            [
                "ObjectCreatedEvent",
                "ObjectWillBeAddedEvent",
                "ObjectAddedEvent",
                "ContainerModifiedEvent",
            ],
        )

        sm.unregisterHandler(record_event, (IObjectCreatedEvent,))
        sm.unregisterHandler(record_event, (IObjectWillBeAddedEvent,))
        sm.unregisterHandler(record_event, (IObjectAddedEvent,))
        sm.unregisterHandler(record_event, (IObjectModifiedEvent,))

    def test_post_to_folder_with_apostrophe_dont_return_500(self):
        response = requests.post(
            self.portal.folder1.absolute_url(),
            headers={"Accept": "application/json"},
            auth=(SITE_OWNER_NAME, SITE_OWNER_PASSWORD),
            json={
                "@type": "Document",
                "id": "mydocument2",
                "title": "My Document 2",
                "text": {
                    "content-type": "text/html",
                    "encoding": "utf8",
                    "data": "<p>example with &#x27;</p>",
                },
            },
        )
        self.assertEqual(201, response.status_code)
        transaction.begin()
        self.assertEqual(
            "<p>example with '</p>", self.portal.folder1.mydocument2.text.raw
        )
        self.assertEqual("<p>example with '</p>", response.json()["text"]["data"])

    def test_post_with_uid_with_manage_portal_permission(self):
        response = requests.post(
            self.portal.folder1.absolute_url(),
            headers={"Accept": "application/json"},
            auth=(SITE_OWNER_NAME, SITE_OWNER_PASSWORD),
            json={
                "@type": "Document",
                "title": "My Document",
                "UID": "a9597fcb108c4985a713329311bdcca0",
            },
        )
        self.assertEqual(201, response.status_code)
        self.assertEqual(response.json()["UID"], "a9597fcb108c4985a713329311bdcca0")

    def test_post_with_uid_without_manage_portal_permission(self):
        user = "test-user-2"
        password = "secret"
        self.portal.acl_users.userFolderAddUser(user, password, ["Contributor"], [])
        transaction.commit()

        response = requests.post(
            self.portal.folder1.absolute_url(),
            headers={"Accept": "application/json"},
            auth=(user, password),
            json={
                "@type": "Document",
                "title": "My Document",
                "UID": "a9597fcb108c4985a713329311bdcca0",
            },
        )
        self.assertEqual(403, response.status_code)
        self.assertEqual(
            response.json()["error"]["message"],
            "Setting UID of an object requires Manage Portal permission",
        )



from plone.restapi.services import Service
from zope.interface import alsoProvides
from plone.app.iterate.interfaces import ICheckinCheckoutPolicy
from zope.component import getMultiAdapter
from Products.CMFCore.utils import getToolByName

import plone.protect.interfaces


class UpdateWorkingCopy(Service):
    def reply(self):
        # Disable CSRF protection
        if "IDisableCSRFProtection" in dir(plone.protect.interfaces):
            alsoProvides(self.request, plone.protect.interfaces.IDisableCSRFProtection)

        policy = ICheckinCheckoutPolicy(self.context)
        working_copy = policy.getWorkingCopy()
        if not policy.getBaseline():
            # We are in the baseline, get the working copy policy
            policy = ICheckinCheckoutPolicy(working_copy)

        control = getMultiAdapter((working_copy, self.request), name="iterate_control")
        if not control.checkin_allowed():
            pm = getToolByName(self.context, "portal_membership")
            if bool(pm.isAnonymousUser()):
                return self._error(401, "Not authenticated", "Checkin not allowed")
            else:
                return self._error(403, "Not authorized", "Checkin not allowed")

        policy.checkin("")

        return self.reply_no_content()

    def _error(self, status, type, message):
        self.request.response.setStatus(status)
        return {"error": {"type": type, "message": message}}



from plone.app.testing import setRoles
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import SITE_OWNER_PASSWORD
from plone.app.testing import TEST_USER_ID
from plone.restapi.testing import PLONE_RESTAPI_DX_FUNCTIONAL_TESTING
from plone.restapi.testing import RelativeSession
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.utils import safe_unicode

import transaction
import unittest


class TestAddons(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_FUNCTIONAL_TESTING

    def setUp(self):
        self.app = self.layer["app"]
        self.portal = self.layer["portal"]
        self.ps = getToolByName(self.portal, "portal_setup")
        self.portal_url = self.portal.absolute_url()
        setRoles(self.portal, TEST_USER_ID, ["Manager"])

        self.api_session = RelativeSession(self.portal_url, test=self)
        self.api_session.headers.update({"Accept": "application/json"})
        self.api_session.auth = (SITE_OWNER_NAME, SITE_OWNER_PASSWORD)

    def test_get_addon_record(self):
        response = self.api_session.get("/@addons/plone.session")

        self.assertEqual(response.status_code, 200)
        result = response.json()

        self.assertEqual(result["@id"], self.portal_url + "/@addons/plone.session")
        self.assertEqual(result["id"], "plone.session")
        # self.assertEqual(result['is_installed'], False)
        self.assertEqual(result["title"], "Session refresh support")
        self.assertEqual(
            result["description"], "Optional plone.session refresh support."
        )
        self.assertEqual(result["profile_type"], "default")
        self.assertEqual(result["upgrade_info"], {})
        self.assertEqual(result["install_profile_id"], "plone.session:default")

    def test_get_addon_listing(self):
        response = self.api_session.get("/@addons")

        self.assertEqual(response.status_code, 200)
        response = response.json()
        self.assertIn("items", response)

    def test_install_uninstall_addon(self):
        def _get_install_status(self):
            response = self.api_session.get("/@addons/plone.session")
            result = response.json()
            return result["is_installed"]

        # Check to make sure the addon is currently shown as not installed
        self.assertEqual(_get_install_status(self), False)

        response = self.api_session.post("/@addons/plone.session/install")
        self.assertEqual(response.status_code, 204)
        self.assertEqual(safe_unicode(response.content), "")

        # Check to make sure the addon is currently shown as installed
        self.assertEqual(_get_install_status(self), True)

        # Now uninstall the addon
        response = self.api_session.post("/@addons/plone.session/uninstall")
        self.assertEqual(response.status_code, 204)
        self.assertEqual(safe_unicode(response.content), "")

        # Check to make sure the addon is currently shown as not installed
        self.assertEqual(_get_install_status(self), False)

    def test_install_uninstall_addon_with_representation(self):

        # Check to make sure the addon is currently shown as not installed
        response = self.api_session.get("/@addons/plone.session")
        result = response.json()
        self.assertEqual(result["is_installed"], False)

        # Install the addon
        response = self.api_session.post(
            "/@addons/plone.session/install",
            headers={"Prefer": "return=representation"},
        )
        self.assertEqual(response.status_code, 200)
        result = response.json()

        # Check to make sure the addon is currently shown as installed
        session = [a for a in result["items"] if a["id"] == "plone.session"]
        self.assertEqual(len(session), 1)
        self.assertTrue(session[0]["is_installed"])

        # Now uninstall the addon
        response = self.api_session.post(
            "/@addons/plone.session/uninstall",
            headers={"Prefer": "return=representation"},
        )
        self.assertEqual(response.status_code, 200)
        result = response.json()

        # Check to make sure the addon is currently shown as not installed
        session = [a for a in result["items"] if a["id"] == "plone.session"]
        self.assertEqual(len(session), 1)
        self.assertFalse(session[0]["is_installed"])

    def test_upgrade_addon(self):
        def _get_upgrade_info(self):
            response = self.api_session.get("/@addons/plone.restapi")
            result = response.json()
            return result["upgrade_info"]

        # Set need upgrade state
        self.ps.setLastVersionForProfile("plone.restapi:default", "0002")
        transaction.commit()
        self.assertEqual(
            {
                "available": True,
                "hasProfile": True,
                "installedVersion": "0002",
                "newVersion": "0006",
                "required": True,
            },
            _get_upgrade_info(self),
        )

        # Now call the upgrade
        response = self.api_session.post("/@addons/plone.restapi/upgrade")
        self.assertEqual(response.status_code, 204)
        self.assertEqual(safe_unicode(response.content), "")
        self.assertEqual(
            {
                "available": False,
                "hasProfile": True,
                "installedVersion": "0006",
                "newVersion": "0006",
                "required": False,
            },
            _get_upgrade_info(self),
        )

    def test_upgrade_addon_with_representation(self):
        response = self.api_session.get("/@addons/plone.restapi")
        result = response.json()
        last_version = result["upgrade_info"]

        # Set need upgrade state
        self.ps.setLastVersionForProfile("plone.restapi:default", "0002")
        transaction.commit()
        response = self.api_session.get("/@addons/plone.restapi")
        result = response.json()
        self.assertEqual(
            {
                "available": True,
                "hasProfile": True,
                "installedVersion": "0002",
                "newVersion": last_version["newVersion"],
                "required": True,
            },
            result["upgrade_info"],
        )

        # Now call the upgrade
        response = self.api_session.post(
            "/@addons/plone.restapi/upgrade",
            headers={"Prefer": "return=representation"},
        )
        self.assertEqual(response.status_code, 200)
        result = response.json()

        # Check to make sure the addon is at last version
        session = [a for a in result["items"] if a["id"] == "plone.restapi"]
        self.assertEqual(len(session), 1)
        self.assertEqual(last_version, session[0]["upgrade_info"])

    def test_get_only_upgradeables(self):
        def _get_upgrade_info(self):
            response = self.api_session.get("/@addons/plone.restapi")
            result = response.json()
            return result["upgrade_info"]

        # Set need upgrade state
        self.ps.setLastVersionForProfile("plone.restapi:default", "0002")
        transaction.commit()
        self.assertEqual(
            {
                "available": True,
                "hasProfile": True,
                "installedVersion": "0002",
                "newVersion": "0006",
                "required": True,
            },
            _get_upgrade_info(self),
        )

        response = self.api_session.get("/@addons?upgradeable=1")

        self.assertEqual(1, len(response.json()["items"]))
        self.assertEqual("plone.restapi", response.json()["items"][0]["id"])



from plone.restapi.services import Service


class DatabaseGet(Service):
    def reply(self):
        db = self.context._p_jar.db()
        return {
            "@id": f"{self.context.absolute_url()}/@database",
            "cache_length": db.cacheSize(),
            "cache_length_bytes": db.getCacheSizeBytes(),
            "cache_detail_length": db.cacheDetailSize(),
            "cache_size": db.getCacheSize(),
            "database_size": db.objectCount(),
            "db_name": db.getName(),
            "db_size": db.getSize(),
        }



from AccessControl import getSecurityManager
from plone.restapi.deserializer import json_body
from plone.restapi.interfaces import ISerializeToJson
from plone.restapi.services import Service
from Products.CMFCore.permissions import AddPortalMember
from Products.CMFCore.permissions import SetOwnPassword
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.interfaces import ISecuritySchema
from Products.CMFPlone.PasswordResetTool import ExpiredRequestError
from Products.CMFPlone.PasswordResetTool import InvalidRequestError
from Products.CMFPlone.RegistrationTool import get_member_by_login_name
from zope.component import getAdapter
from zope.component import getMultiAdapter
from zope.component import queryMultiAdapter
from zope.component.hooks import getSite
from zope.interface import alsoProvides
from zope.interface import implementer
from zope.publisher.interfaces import IPublishTraverse

import plone.protect.interfaces


@implementer(IPublishTraverse)
class UsersPost(Service):
    """Creates a new user."""

    def __init__(self, context, request):
        super().__init__(context, request)
        self.params = []

    def publishTraverse(self, request, name):
        # Consume any path segments after /@users as parameters
        self.params.append(name)
        return self

    def validate_input_data(self, portal, original_data):
        """Returns a tuple of (required_fields, allowed_fields)"""
        security = getAdapter(portal, ISecuritySchema)

        # remove data we don't want to check for
        data = {}
        for key in ["username", "email", "password", "roles", "sendPasswordReset"]:
            if key in original_data:
                data[key] = original_data[key]

        required = ["email"]
        allowed = ["email"]

        if not security.use_email_as_login:
            required.append("username")
            allowed.append("username")

        if self.can_manage_users:
            allowed.append("password")
            allowed.append("sendPasswordReset")
            allowed.append("roles")
        else:
            if security.enable_user_pwd_choice:
                allowed.append("password")
                required.append("password")

        # check input data
        for fieldname in required:
            if not data.get(fieldname, None):
                self.add_field_error(fieldname, f"Property '{fieldname}' is required.")
        for fieldname in data:
            if fieldname not in allowed:
                self.add_field_error(
                    fieldname, f"Property '{fieldname}' is not allowed."
                )

        password = data.get("password")
        send_password_reset = data.get("sendPasswordReset")
        if self.can_manage_users:
            if password is None and send_password_reset is None:
                self.add_field_error(
                    "sendPasswordReset",
                    "You have to either send a password or sendPasswordReset.",
                )
            if password and send_password_reset:
                self.add_field_error(
                    "sendPasswordReset",
                    "You can't send both password and sendPasswordReset.",
                )

    def add_field_error(self, field, message):
        self.errors.append({"field": field, "message": message})

    def errors_to_string(self):
        return " ".join([error["message"] for error in self.errors])

    def reply(self):
        # Disable CSRF protection
        if "IDisableCSRFProtection" in dir(plone.protect.interfaces):
            alsoProvides(self.request, plone.protect.interfaces.IDisableCSRFProtection)

        portal = getSite()

        # validate important data
        data = json_body(self.request)
        self.errors = []
        self.validate_input_data(portal, data)
        security = getAdapter(self.context, ISecuritySchema)
        registration = getToolByName(self.context, "portal_registration")

        general_usage_error = (
            "Either post to @users to create a user or use "
            "@users/<username>/reset-password to update the password."
        )
        if len(self.params) not in [0, 2]:
            raise Exception(general_usage_error)

        if len(self.params) == 2:
            if self.params[1] == "reset-password":
                return self.update_password(data)
            raise Exception("Unknown Endpoint @users/%s/%s" % self.params)

        # Add a portal member
        if not self.can_add_member:
            return self._error(403, "Forbidden", "You need AddPortalMember permission.")

        if self.errors:
            self.request.response.setStatus(400)
            return dict(
                error=dict(
                    type="WrongParameterError",
                    message=f"Error in fields. {self.errors_to_string()}",
                    errors=self.errors,
                )
            )

        username = data.pop("username", None)
        email = data.pop("email", None)
        password = data.pop("password", None)
        roles = data.pop("roles", ["Member"])
        send_password_reset = data.pop("sendPasswordReset", None)
        properties = data

        user_id_login_name_data = {
            "username": username,
            "email": email,
            "fullname": data.get("fullname", ""),
        }

        register_view = getMultiAdapter((self.context, self.request), name="register")

        register_view.generate_user_id(user_id_login_name_data)
        register_view.generate_login_name(user_id_login_name_data)

        user_id = user_id_login_name_data.get("user_id", data.get("username"))
        login_name = user_id_login_name_data.get("login_name", data.get("username"))

        username = user_id
        properties["username"] = user_id

        properties["email"] = email

        if not self.can_manage_users and not security.enable_user_pwd_choice:
            send_password_reset = True
        if send_password_reset:
            password = registration.generatePassword()
        # Create user
        try:
            registration = getToolByName(portal, "portal_registration")
            user = registration.addMember(
                username, password, roles, properties=properties
            )
        except ValueError as e:
            self.request.response.setStatus(400)
            return dict(error=dict(type="MissingParameterError", message=str(e)))

        if user_id != login_name:
            # The user id differs from the login name.  Set the login
            # name correctly.
            pas = getToolByName(self.context, "acl_users")
            pas.updateLoginName(user_id, login_name)

        if send_password_reset:
            registration.registeredNotify(username)
        self.request.response.setStatus(201)
        self.request.response.setHeader(
            "Location", portal.absolute_url() + "/@users/" + username
        )
        serializer = queryMultiAdapter((user, self.request), ISerializeToJson)
        return serializer()

    def _get_user(self, user_id):
        portal = getSite()
        portal_membership = getToolByName(portal, "portal_membership")
        return portal_membership.getMemberById(user_id)

    def _get_user_by_login_name(self, user_id):
        return get_member_by_login_name(self.context, user_id, raise_exceptions=False)

    def _error(self, status, type, message):
        self.request.response.setStatus(status)
        return {"error": {"type": type, "message": message}}

    @property
    def can_manage_users(self):
        sm = getSecurityManager()
        return sm.checkPermission("plone.app.controlpanel.UsersAndGroups", self.context)

    @property
    def can_set_own_password(self):
        sm = getSecurityManager()
        return sm.checkPermission(SetOwnPassword, self.context)

    @property
    def can_add_member(self):
        sm = getSecurityManager()
        return sm.checkPermission(AddPortalMember, self.context)

    def update_password(self, data):
        username = self.params[0]
        target_user = self._get_user_by_login_name(username)
        reset_token = data.get("reset_token", None)
        old_password = data.get("old_password", None)
        new_password = data.get("new_password", None)

        pas = getToolByName(self.context, "acl_users")
        mt = getToolByName(self.context, "portal_membership")
        pwt = getToolByName(self.context, "portal_password_reset")

        if target_user is None:
            self.request.response.setStatus(404)
            return

        # Send password reset mail
        if list(data) == []:
            registration_tool = getToolByName(self.context, "portal_registration")
            registration_tool.mailPassword(username, self.request)
            return

        if reset_token and old_password:
            return self._error(
                400,
                "Invalid parameters",
                "You can't use 'reset_token' and 'old_password' together.",
            )
        if reset_token and not new_password:
            return self._error(
                400,
                "Invalid parameters",
                "If you pass 'reset_token' you have to pass 'new_password'",
            )
        if old_password and not new_password:
            return self._error(
                400,
                "Invalid parameters",
                "If you pass 'old_password' you have to pass 'new_password'",
            )

        # Reset the password with a reset token
        if reset_token:
            try:
                pwt.resetPassword(username, reset_token, new_password)
            except InvalidRequestError:
                return self._error(
                    403, "Unknown Token", "The reset_token is unknown/not valid."
                )
            except ExpiredRequestError:
                return self._error(403, "Expired Token", "The reset_token is expired.")
            return

        # set the new password by giving the old password
        if old_password:
            if not (self.can_manage_users or self.can_set_own_password):
                return self._error(
                    403,
                    "Not allowed",
                    "You can't set a password without " "a password reset token.",
                )
            authenticated_user_id = mt.getAuthenticatedMember().getId()
            if username != authenticated_user_id:
                return self._error(
                    403,
                    "Wrong user",
                    (
                        "You need to be logged in as the user '%s' to set "
                        "the password."
                    )
                    % username,
                )

            check_password_auth = pas.authenticate(
                username, old_password.encode("utf-8"), self.request
            )
            if not check_password_auth:
                return self._error(
                    403,
                    "Wrong password",
                    "The password passed as 'old_password' " "is wrong.",
                )
            mt.setPassword(new_password)
            return

        return self._error(
            400,
            "Invalid parameters",
            "See the user endpoint documentation for the " "valid parameters.",
        )



from plone.restapi.services import Service
from plone.restapi.services.addons.addons import Addons
from zope.component import getMultiAdapter
from zope.interface import alsoProvides
from zope.interface import implementer
from zope.publisher.interfaces import IPublishTraverse

import logging
import plone


logger = logging.getLogger("Plone")


@implementer(IPublishTraverse)
class AddonsPost(Service):
    """Performs install/upgrade/uninstall functions on an addon."""

    def __init__(self, context, request):
        super().__init__(context, request)
        self.params = []
        self.errors = {}
        self.addons = Addons(context, request)

    def publishTraverse(self, request, name):
        # Consume any path segments after /@addons as parameters
        self.params.append(name)
        return self

    def reply(self):
        addon, action = self.params

        # Disable CSRF protection
        if "IDisableCSRFProtection" in dir(plone.protect.interfaces):
            alsoProvides(self.request, plone.protect.interfaces.IDisableCSRFProtection)

        if action == "install":
            result = self.addons.install_product(addon)
        elif action == "uninstall":
            result = self.addons.uninstall_product(addon)
        elif action == "upgrade":
            result = self.addons.upgrade_product(addon)
        else:
            raise Exception(f"Unknown action {action}")

        prefer = self.request.getHeader("Prefer")
        if prefer == "return=representation":
            control_panel = getMultiAdapter(
                (self.context, self.request), name="prefs_install_products_form"
            )
            all_addons = control_panel.get_addons()

            result = {"items": {"@id": f"{self.context.absolute_url()}/@addons"}}
            addons_data = []
            for a in all_addons.values():
                addons_data.append(self.addons.serializeAddon(a))
            result["items"] = addons_data

            self.request.response.setStatus(200)
            return result
        self.request.response.setStatus(204)



from plone.dexterity.interfaces import IDexterityFTI
from plone.dexterity.interfaces import IDexterityItem
from plone.dexterity.utils import iterSchemata
from plone.restapi.behaviors import IBlocks
from plone.restapi.interfaces import IBlockFieldSerializationTransformer
from plone.restapi.interfaces import IFieldSerializer
from plone.restapi.testing import PLONE_RESTAPI_DX_INTEGRATION_TESTING
from plone.uuid.interfaces import IUUID
from z3c.form.interfaces import IDataManager
from zope.component import adapter
from zope.component import getGlobalSiteManager
from zope.component import getMultiAdapter
from zope.component import provideSubscriptionAdapter
from zope.component import queryUtility
from zope.interface import implementer
from zope.publisher.interfaces.browser import IBrowserRequest

import unittest


class TestBlocksSerializer(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_INTEGRATION_TESTING

    def setUp(self):
        self.app = self.layer["app"]
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]

        fti = queryUtility(IDexterityFTI, name="Document")
        behavior_list = [a for a in fti.behaviors]
        behavior_list.append("volto.blocks")
        fti.behaviors = tuple(behavior_list)

        self.portal.invokeFactory("Document", id="doc1")
        self.image = self.portal[
            self.portal.invokeFactory("Image", id="image-1", title="Target image")
        ]

    def serialize(self, context, blocks):
        fieldname = "blocks"
        field = None
        for schema in iterSchemata(context):
            if fieldname in schema:
                field = schema.get(fieldname)
                break
        if field is None:
            raise ValueError("blocks field not in the schema of %s" % context)
        dm = getMultiAdapter((context, field), IDataManager)
        dm.set(blocks)
        serializer = getMultiAdapter((field, context, self.request), IFieldSerializer)
        return serializer()

    def test_register_serializer(self):
        @adapter(IBlocks, IBrowserRequest)
        @implementer(IBlockFieldSerializationTransformer)
        class TestAdapterA:
            order = 10
            block_type = "test_multi"

            def __init__(self, context, request):
                self.context = context
                self.request = request

            def __call__(self, value):
                self.context._handler_called_a = True

                value["value"] = value["value"].replace("a", "b")

                return value

        @adapter(IBlocks, IBrowserRequest)
        @implementer(IBlockFieldSerializationTransformer)
        class TestAdapterB:
            order = 11
            block_type = "test_multi"

            def __init__(self, context, request):
                self.context = context
                self.request = request

            def __call__(self, value):
                self.context._handler_called_b = True

                value["value"] = value["value"].replace("b", "c")

                return value

        provideSubscriptionAdapter(TestAdapterA, (IDexterityItem, IBrowserRequest))
        provideSubscriptionAdapter(TestAdapterB, (IDexterityItem, IBrowserRequest))
        value = self.serialize(
            context=self.portal.doc1,
            blocks={"123": {"@type": "test_multi", "value": "a"}},
        )
        self.assertEqual(value["123"]["value"], "c")

        sm = getGlobalSiteManager()
        sm.adapters.unsubscribe(
            (IDexterityItem, IBrowserRequest),
            IBlockFieldSerializationTransformer,
            TestAdapterA,
        )
        sm.adapters.unsubscribe(
            (IDexterityItem, IBrowserRequest),
            IBlockFieldSerializationTransformer,
            TestAdapterB,
        )

    def test_disabled_serializer(self):
        @implementer(IBlockFieldSerializationTransformer)
        @adapter(IBlocks, IBrowserRequest)
        class TestAdapter:
            order = 10
            block_type = "test"
            disabled = True

            def __init__(self, context, request):
                self.context = context
                self.request = request

            def __call__(self, value):
                self.context._handler_called = True

                value["value"] = "changed: {}".format(value["value"])

                return value

        provideSubscriptionAdapter(
            TestAdapter,
            (IDexterityItem, IBrowserRequest),
        )
        value = self.serialize(
            context=self.portal.doc1,
            blocks={"123": {"@type": "test", "value": "text"}},
        )

        assert not getattr(self.portal.doc1, "_handler_called", False)
        self.assertEqual(value["123"]["value"], "text")

        sm = getGlobalSiteManager()
        sm.adapters.unsubscribe(
            (IDexterityItem, IBrowserRequest),
            IBlockFieldSerializationTransformer,
            TestAdapter,
        )

    def test_serialize_blocks_smart_href_array_volto_object_browser(self):
        doc_uid = IUUID(self.portal.doc1)
        value = self.serialize(
            context=self.portal.doc1,
            blocks={
                "123": {
                    "@type": "foo",
                    "href": [{"@id": f"../resolveuid/{doc_uid}"}],
                }
            },
        )

        self.assertEqual(
            value["123"]["href"][0]["@id"], self.portal.doc1.absolute_url()
        )

    def test_serialize_blocks_smart_href_array(self):
        doc_uid = IUUID(self.portal.doc1)
        value = self.serialize(
            context=self.portal.doc1,
            blocks={"123": {"@type": "foo", "href": [f"../resolveuid/{doc_uid}"]}},
        )

        self.assertEqual(value["123"]["href"][0], self.portal.doc1.absolute_url())

    def test_serialize_subblocks_transformers(self):
        # use the href smart field transformer for testing subblocks transformers
        doc_uid = IUUID(self.portal.doc1)
        subblock = {"@type": "foo", "href": [f"../resolveuid/{doc_uid}"]}
        value = self.serialize(
            context=self.portal.doc1,
            blocks={
                "1": {
                    "@type": "columns_block",
                    "data": {
                        "blocks": {"2": {"@type": "tabs", "blocks": {"3": subblock}}}
                    },
                }
            },
        )

        href = value["1"]["data"]["blocks"]["2"]["blocks"]["3"]["href"]

        self.assertEqual(href[0], self.portal.doc1.absolute_url())

    def test_internal_link_serializer(self):
        doc_uid = IUUID(self.portal["doc1"])
        resolve_uid_link = {
            "@id": f"../resolveuid/{doc_uid}",
            "title": "Welcome to Plone",
        }
        blocks = {
            "2caef9e6-93ff-4edf-896f-8c16654a9923": {
                "@type": "slate",
                "plaintext": "this is a slate link inside some text",
                "value": [
                    {
                        "children": [
                            {"text": "this is a "},
                            {
                                "children": [
                                    {"text": ""},
                                    {
                                        "children": [{"text": "slate link"}],
                                        "data": {
                                            "link": {
                                                "internal": {
                                                    "internal_link": [resolve_uid_link]
                                                }
                                            }
                                        },
                                        "type": "a",
                                    },
                                    {"text": ""},
                                ],
                                "type": "strong",
                            },
                            {"text": " inside some text"},
                        ],
                        "type": "p",
                    }
                ],
            },
            "6b2be2e6-9857-4bcc-a21a-29c0449e1c68": {"@type": "title"},
        }

        res = self.serialize(
            context=self.portal["doc1"],
            blocks=blocks,
        )

        value = res["2caef9e6-93ff-4edf-896f-8c16654a9923"]["value"]
        link = value[0]["children"][1]["children"][1]
        resolve_link = link["data"]["link"]["internal"]["internal_link"][0]["@id"]

        self.assertTrue(resolve_link == self.portal.absolute_url() + "/doc1")

    def test_simple_link_serializer(self):
        doc_uid = IUUID(self.portal["doc1"])
        resolve_uid_link = f"../resolveuid/{doc_uid}"

        blocks = {
            "abc": {
                "@type": "slate",
                "plaintext": "Frontpage content here",
                "value": [
                    {
                        "children": [
                            {"text": "Frontpage "},
                            {
                                "children": [{"text": "content "}],
                                "data": {
                                    "url": resolve_uid_link,
                                },
                                "type": "link",
                            },
                            {"text": "here"},
                        ],
                        "type": "h2",
                    }
                ],
            }
        }
        res = self.serialize(
            context=self.portal["doc1"],
            blocks=blocks,
        )
        value = res["abc"]["value"]
        link = value[0]["children"][1]["data"]["url"]
        self.assertTrue(link, self.portal.absolute_url() + "/doc1")



class APIError(Exception):
    """Base class for exceptions raised by plone.restapi."""


class DeserializationError(Exception):
    """An error happened during deserialization of content."""

    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return repr(self.msg)


class QueryParsingError(Exception):
    """An error happened while parsing a search query."""



from DateTime import DateTime
from plone.restapi.exceptions import QueryParsingError
from plone.restapi.search.query import BaseIndexQueryParser
from plone.restapi.search.query import BooleanIndexQueryParser
from plone.restapi.search.query import DateIndexQueryParser
from plone.restapi.search.query import DateRangeIndexQueryParser
from plone.restapi.search.query import ExtendedPathIndexQueryParser
from plone.restapi.search.query import FieldIndexQueryParser
from plone.restapi.search.query import KeywordIndexQueryParser
from plone.restapi.search.query import UUIDIndexQueryParser
from plone.restapi.search.query import ZCTextIndexQueryParser

import unittest


class TestBaseIndexQueryParser(unittest.TestCase):
    def test_casts_simple_query_to_string(self):
        qp = BaseIndexQueryParser()
        self.assertEqual("42", qp.parse(42))

    def test_casts_complex_query_values_to_string(self):
        qp = BaseIndexQueryParser()
        self.assertEqual({"query": "42"}, qp.parse({"query": 42}))

    def test_casts_query_values_in_sequences(self):
        qp = BaseIndexQueryParser()
        self.assertEqual(["23", "42"], qp.parse([23, 42]))
        self.assertEqual(["23", "42"], qp.parse((23, 42)))

    def test_raises_on_missing_query_key_for_complex_queries(self):
        qp = BaseIndexQueryParser()
        with self.assertRaises(QueryParsingError):
            qp.parse({})


class TestZCTextIndexQueryParser(unittest.TestCase):
    def test_casts_simple_query_to_string(self):
        qp = ZCTextIndexQueryParser()
        self.assertEqual("42", qp.parse(42))

    def test_casts_complex_query_values_to_string(self):
        qp = ZCTextIndexQueryParser()
        self.assertEqual({"query": "42"}, qp.parse({"query": 42}))


class TestKeywordIndexQueryParser(unittest.TestCase):
    def test_returns_simple_query_unchanged(self):
        qp = KeywordIndexQueryParser()
        self.assertEqual("keyword", qp.parse("keyword"))
        self.assertEqual(42, qp.parse(42))

    def test_returns_complex_query_values_unchanged(self):
        qp = KeywordIndexQueryParser()
        self.assertEqual({"query": "keyword"}, qp.parse({"query": "keyword"}))
        self.assertEqual({"query": 42}, qp.parse({"query": 42}))

    def test_casts_operator_option_to_string(self):
        qp = KeywordIndexQueryParser()
        self.assertEqual(
            {"operator": "42", "query": "keyword"},
            qp.parse({"operator": 42, "query": "keyword"}),
        )

    def test_casts_range_option_to_string(self):
        qp = KeywordIndexQueryParser()
        self.assertEqual(
            {"range": "42", "query": "keyword"},
            qp.parse({"range": 42, "query": "keyword"}),
        )


class TestBooleanIndexQueryParser(unittest.TestCase):
    def test_casts_simple_query_to_boolean(self):
        qp = BooleanIndexQueryParser()
        self.assertEqual(True, qp.parse("True"))
        self.assertEqual(True, qp.parse("true"))
        self.assertEqual(True, qp.parse("1"))
        self.assertEqual(False, qp.parse("False"))
        self.assertEqual(False, qp.parse("false"))
        self.assertEqual(False, qp.parse("0"))

    def test_casts_complex_query_values_to_boolean(self):
        qp = BooleanIndexQueryParser()
        self.assertEqual({"query": True}, qp.parse({"query": "True"}))
        self.assertEqual({"query": True}, qp.parse({"query": "true"}))
        self.assertEqual({"query": True}, qp.parse({"query": "1"}))
        self.assertEqual({"query": False}, qp.parse({"query": "False"}))
        self.assertEqual({"query": False}, qp.parse({"query": "false"}))
        self.assertEqual({"query": False}, qp.parse({"query": "0"}))

    def test_raises_for_invalid_query_type(self):
        qp = BooleanIndexQueryParser()
        with self.assertRaises(QueryParsingError):
            qp.parse(42)


class TestFieldIndexQueryParser(unittest.TestCase):
    def test_returns_simple_query_unchanged(self):
        qp = FieldIndexQueryParser()
        self.assertEqual("foo", qp.parse("foo"))
        self.assertEqual(42, qp.parse(42))

    def test_returns_complex_query_values_unchanged(self):
        qp = FieldIndexQueryParser()
        self.assertEqual({"query": "foo"}, qp.parse({"query": "foo"}))
        self.assertEqual({"query": 42}, qp.parse({"query": 42}))

    def test_casts_range_option_to_string(self):
        qp = FieldIndexQueryParser()
        self.assertEqual(
            {"range": "42", "query": "/path"}, qp.parse({"range": 42, "query": "/path"})
        )


class TestExtendedPathIndexQueryParser(unittest.TestCase):
    def test_casts_simple_query_to_string(self):
        qp = ExtendedPathIndexQueryParser()
        self.assertEqual("42", qp.parse(42))

    def test_casts_complex_query_values_to_string(self):
        qp = ExtendedPathIndexQueryParser()
        self.assertEqual({"query": "42"}, qp.parse({"query": 42}))

    def test_casts_level_option_to_int(self):
        qp = ExtendedPathIndexQueryParser()
        self.assertEqual(
            {"level": 3, "query": "/path"}, qp.parse({"level": "3", "query": "/path"})
        )

    def test_casts_operator_option_to_string(self):
        qp = ExtendedPathIndexQueryParser()
        self.assertEqual(
            {"operator": "42", "query": "/path"},
            qp.parse({"operator": 42, "query": "/path"}),
        )

    def test_casts_depth_option_to_int(self):
        qp = ExtendedPathIndexQueryParser()
        self.assertEqual(
            {"depth": 3, "query": "/path"}, qp.parse({"depth": "3", "query": "/path"})
        )

    def test_casts_navtree_option_to_int(self):
        qp = ExtendedPathIndexQueryParser()
        self.assertEqual(
            {"navtree": False, "query": "/path"},
            qp.parse({"navtree": 0, "query": "/path"}),
        )
        self.assertEqual(
            {"navtree": True, "query": "/path"},
            qp.parse({"navtree": 1, "query": "/path"}),
        )

    def test_casts_navtree_start_option_to_int(self):
        qp = ExtendedPathIndexQueryParser()
        self.assertEqual(
            {"navtree_start": 42, "query": "/path"},
            qp.parse({"navtree_start": "42", "query": "/path"}),
        )


class TestDateIndexQueryParser(unittest.TestCase):
    def test_casts_simple_query_to_zope_date_time(self):
        qp = DateIndexQueryParser()
        self.assertEqual(DateTime("2016/12/24 00:00:00 GMT+0"), qp.parse("2016-12-24"))

    def test_casts_complex_query_values_to_zope_date_time(self):
        qp = DateIndexQueryParser()
        self.assertEqual(
            {"query": DateTime("2016/12/24 00:00:00 GMT+0")},
            qp.parse({"query": "2016-12-24"}),
        )

    def test_casts_range_option_to_string(self):
        qp = DateIndexQueryParser()
        self.assertEqual(
            {"range": "42", "query": DateTime("2016/12/24 00:00:00 GMT+0")},
            qp.parse({"range": 42, "query": "2016-12-24"}),
        )


class TestDateRangeIndexQueryParser(unittest.TestCase):
    def test_casts_simple_query_to_zope_date_time(self):
        qp = DateRangeIndexQueryParser()
        self.assertEqual(DateTime("2016/12/24 00:00:00 GMT+0"), qp.parse("2016-12-24"))

    def test_casts_complex_query_values_to_zope_date_time(self):
        qp = DateRangeIndexQueryParser()
        self.assertEqual(
            {"query": DateTime("2016/12/24 00:00:00 GMT+0")},
            qp.parse({"query": "2016-12-24"}),
        )


class TestUUIDIndexQueryParser(unittest.TestCase):
    def test_casts_simple_query_to_string(self):
        qp = UUIDIndexQueryParser()
        self.assertEqual("42", qp.parse(42))

    def test_casts_complex_query_values_to_string(self):
        qp = UUIDIndexQueryParser()
        self.assertEqual({"query": "42"}, qp.parse({"query": 42}))

    def test_casts_range_option_to_string(self):
        qp = UUIDIndexQueryParser()
        self.assertEqual(
            {"range": "42", "query": "<UID>"}, qp.parse({"range": 42, "query": "<UID>"})
        )



from plone.restapi.interfaces import IPloneRestapiLayer
from plone.restapi.services import Service
from plone.restapi.types.utils import delete_field
from plone.restapi.types.utils import delete_fieldset
from zExceptions import BadRequest
from zope.component import queryMultiAdapter
from zope.interface import alsoProvides
from zope.interface import implementer
from zope.interface import noLongerProvides
from zope.publisher.interfaces import IPublishTraverse
import plone.protect.interfaces


@implementer(IPublishTraverse)
class TypesDelete(Service):
    """Deletes a field/fieldset from content type"""

    def __init__(self, context, request):
        super().__init__(context, request)
        self.params = []

    def publishTraverse(self, request, name):
        # Treat any path segments after /@types as parameters
        self.params.append(name)
        return self

    def reply(self):
        if not self.params:
            raise BadRequest("Missing parameter typename")
        if len(self.params) < 2:
            raise BadRequest("Missing parameter fieldname")
        if len(self.params) > 2:
            raise BadRequest("Too many parameters")

        # Disable CSRF protection
        if "IDisableCSRFProtection" in dir(plone.protect.interfaces):
            alsoProvides(self.request, plone.protect.interfaces.IDisableCSRFProtection)

        # Make sure we don't get the right dexterity-types adapter
        if IPloneRestapiLayer.providedBy(self.request):
            noLongerProvides(self.request, IPloneRestapiLayer)

        context = queryMultiAdapter(
            (self.context, self.request), name="dexterity-types"
        )

        # Get content type SchemaContext
        name = self.params.pop(0)
        context = context.publishTraverse(self.request, name)

        name = self.params.pop(0)
        try:
            context.publishTraverse(self.request, name)
        except AttributeError:
            delete_fieldset(context, self.request, name)
        else:
            delete_field(context, self.request, name)

        return self.reply_no_content()



from plone.app.i18n.locales.languages import ContentLanguages
from plone.app.i18n.locales.languages import MetadataLanguages
from plone.i18n.locales.languages import _combinedlanguagelist
from plone.i18n.locales.languages import _languagelist


class ModifiableLanguages:
    """Mixin for the `IModifiableLanguageAvailability` based local utilities
    that makes the getLanguages and getLanguageListing methods
    (queried by portal_languages) also respect *modifiable* language
    availability.

    These methods would otherwise be inherited from LanguageAvailability
    (plone.i18n), and we override those here with variants that *do* respect
    modifiable language availability.

    This is so we can work with a limited set of languages in tests in order
    to avoid excessive language lists in response dumps included in docs.
    """

    def getLanguages(self, combined=False):
        """Return a sequence of Language objects for available languages."""
        languages = _languagelist.copy()
        if combined:
            languages.update(_combinedlanguagelist.copy())

        available = self.getAvailableLanguages(combined=combined)
        languages = {k: v for k, v in languages.items() if k in available}

        return languages

    def getLanguageListing(self, combined=False):
        """Return a sequence of language code and language name tuples."""
        languages = _languagelist.copy()
        if combined:
            languages.update(_combinedlanguagelist.copy())

        available = self.getAvailableLanguages(combined=combined)
        languages = {k: v for k, v in languages.items() if k in available}

        return [(code, languages[code]["name"]) for code in languages]


class ModifiableContentLanguages(ModifiableLanguages, ContentLanguages):
    """Overrides /plone/plone_app_content_languages with a variant that
    fully respects modifiable language availability.
    """


class ModifiableMetadataLanguages(ModifiableLanguages, MetadataLanguages):
    """Overrides /plone/plone_app_metadata_languages with a variant that
    fully respects modifiable language availability.
    """



from plone.dexterity.utils import iterSchemata
from plone.restapi.interfaces import ISerializeToJson
from plone.restapi.services import Service
from zope.component import getMultiAdapter
from zope.interface import implementer
from zope.publisher.interfaces import IPublishTraverse
from zope.schema import getFieldsInOrder
from zope.schema.interfaces import IIterableSource
from zope.schema.interfaces import ISource


@implementer(IPublishTraverse)
class SourcesGet(Service):
    def __init__(self, context, request):
        super().__init__(context, request)
        self.params = []

    def publishTraverse(self, request, name):
        # Treat any path segments after /@sources as parameters
        self.params.append(name)
        return self

    def _error(self, status, type, message):
        self.request.response.setStatus(status)
        return {"error": {"type": type, "message": message}}

    def reply(self):
        if len(self.params) != 1:
            return self._error(
                400, "Bad Request", "Must supply exactly one path parameter (fieldname)"
            )
        fieldname = self.params[0]

        field = get_field_by_name(fieldname, self.context)
        if field is None:
            return self._error(404, "Not Found", "No such field: %r" % fieldname)
        bound_field = field.bind(self.context)

        source = bound_field.source
        if not ISource.providedBy(source):
            return self._error(
                404, "Not Found", "Field %r does not have a source" % fieldname
            )

        if not IIterableSource.providedBy(source):
            return self._error(
                400, "Bad Request", "Source for field %r is not iterable. " % fieldname
            )

        serializer = getMultiAdapter((source, self.request), interface=ISerializeToJson)
        return serializer(f"{self.context.absolute_url()}/@sources/{fieldname}")


def get_field_by_name(fieldname, context):
    schemata = iterSchemata(context)
    for schema in schemata:
        fields = getFieldsInOrder(schema)
        for fn, field in fields:
            if fn == fieldname:
                return field



from AccessControl.SecurityManagement import getSecurityManager
from plone.rest import Service as RestService
from plone.restapi.permissions import UseRESTAPI
from zExceptions import Unauthorized

import json


_no_content_marker = object()


class Service(RestService):
    """Base class for Plone REST API services"""

    content_type = "application/json"

    def render(self):
        self.check_permission()
        content = self.reply()
        if content is not _no_content_marker:
            self.request.response.setHeader("Content-Type", self.content_type)
            return json.dumps(
                content, indent=2, sort_keys=True, separators=(", ", ": ")
            )

    def check_permission(self):
        sm = getSecurityManager()
        if not sm.checkPermission(UseRESTAPI, self):
            raise Unauthorized("Missing %r permission" % UseRESTAPI)

    def reply(self):
        """Process the request and return a JSON serializable data structure or
        the no content marker if the response body should be empty.
        """
        return _no_content_marker

    def reply_no_content(self, status=204):
        self.request.response.setStatus(status)
        return _no_content_marker



from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from plone.restapi.testing import PLONE_RESTAPI_DX_INTEGRATION_TESTING
from plone.restapi.upgrades.ordering import (
    ensure_child_ordering_object_ids_are_native_strings,
)

import unittest


class TestUpgradeOrdering(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]

        self.folder = self.portal[
            self.portal.invokeFactory("Folder", id="folder1", title="Folder")
        ]
        for x in range(1, 4):
            self.folder.invokeFactory(
                "Document", id="doc" + str(x), title="Test doc " + str(x)
            )

    def test_upgrade_ensure_child_ordering_object_ids_are_native_strings(self):
        ordering = self.folder.getOrdering()

        # use incorrect type for ordering, results in mixed type ordering ids
        # on folder
        ordering.moveObjectsToBottom(["doc1"])

        ensure_child_ordering_object_ids_are_native_strings(self.folder)

        self.assertEqual(
            [
                "doc2",
                "doc3",
                "doc1",
            ],
            self.folder.objectIds(),  # noqa
        )

        # upgrade helper should ensure bytestring ids in python2 and do nothing
        # on python3
        for id_ in self.folder.objectIds():
            self.assertIsInstance(id_, str)

    def test_upgrade_can_be_called_with_nonetype(self):
        ensure_child_ordering_object_ids_are_native_strings(None)

    def test_upgrade_can_be_called_with_not_annotatable(self):
        ensure_child_ordering_object_ids_are_native_strings(object())

    def test_upgrade_view(self):
        ordering = self.folder.getOrdering()
        # use incorrect type for ordering, results in mixed type ordering ids
        # on folder
        ordering.moveObjectsToBottom(["doc1"])

        setRoles(self.portal, TEST_USER_ID, ["Manager"])
        view = self.portal.restrictedTraverse("@@plone-restapi-upgrade-fix-ordering")
        view()

        self.assertEqual(
            [
                "doc2",
                "doc3",
                "doc1",
            ],
            self.folder.objectIds(),  # noqa
        )

        # upgrade helper should ensure bytestring ids in python2 and do nothing
        # on python3
        for id_ in self.folder.objectIds():
            self.assertIsInstance(id_, str)



from Acquisition import aq_inner
from Acquisition import aq_parent
from plone.restapi.pas.plugin import JWTAuthenticationPlugin
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.interfaces import INonInstallable
from Products.PluggableAuthService.interfaces.authservice import (
    IPluggableAuthService,
)  # noqa: E501
from zope.component.hooks import getSite
from zope.interface import implementer


@implementer(INonInstallable)
class HiddenProfiles:
    def getNonInstallableProfiles(self):  # pragma: no cover
        """Do not show on Plone's list of installable profiles."""
        return [
            "plone.restapi:blocks",
            "plone.restapi:performance",
            "plone.restapi:testing",
            "plone.restapi:testing-workflows",
            "plone.restapi:uninstall",
        ]

    def getNonInstallableProducts(self):  # pragma: no cover
        """Do not show on Plone's list of installable products.

        This method is only used in Plone 5.1+.
        """
        return ["plone.restapi.upgrades"]


def install_pas_plugin(context):
    uf_parent = aq_inner(context)
    while True:
        uf = getToolByName(uf_parent, "acl_users")
        if IPluggableAuthService.providedBy(uf) and "jwt_auth" not in uf:
            plugin = JWTAuthenticationPlugin("jwt_auth")
            uf._setObject(plugin.getId(), plugin)
            plugin = uf["jwt_auth"]
            plugin.manage_activateInterfaces(
                ["IAuthenticationPlugin", "IExtractionPlugin"]
            )
        if uf_parent is uf_parent.getPhysicalRoot():
            break
        uf_parent = aq_parent(uf_parent)


def post_install_default(context):
    """Post install of default profile"""
    site = getSite()
    install_pas_plugin(site)



# pylint: disable=E1002
# E1002: Use of super on an old style class
from plone.app.contenttypes.testing import PLONE_APP_CONTENTTYPES_FIXTURE
from plone.app.i18n.locales.interfaces import IContentLanguages
from plone.app.i18n.locales.interfaces import IMetadataLanguages
from plone.app.iterate.testing import PLONEAPPITERATEDEX_FIXTURE
from plone.app.testing import applyProfile
from plone.app.testing import FunctionalTesting
from plone.app.testing import IntegrationTesting
from plone.app.testing import login
from plone.app.testing import PloneSandboxLayer
from plone.app.testing import quickInstallProduct
from plone.app.testing import setRoles
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import SITE_OWNER_PASSWORD
from plone.app.testing import TEST_USER_ID
from plone.i18n.interfaces import ILanguageSchema
from plone.registry.interfaces import IRegistry
from plone.restapi.tests.dxtypes import INDEXES as DX_TYPES_INDEXES
from plone.restapi.tests.helpers import add_catalog_indexes
from plone.testing import z2
from plone.testing.layer import Layer
from plone.uuid.interfaces import IUUIDGenerator
from Products.CMFCore.utils import getToolByName
from requests.exceptions import ConnectionError
from urllib.parse import urljoin
from urllib.parse import urlparse
from zope.component import getGlobalSiteManager
from zope.component import getUtility
from zope.configuration import xmlconfig
from zope.interface import implementer

import collective.MockMailHost
import os
import re
import requests
import time


ENABLED_LANGUAGES = ["de", "en", "es", "fr"]


def set_available_languages():
    """Limit available languages to a small set.

    We're doing this to avoid excessive language lists in dumped responses
    for docs. Depends on our own ModifiableLanguages components
    (see plone.restapi:testing profile).
    """
    getUtility(IContentLanguages).setAvailableLanguages(ENABLED_LANGUAGES)
    getUtility(IMetadataLanguages).setAvailableLanguages(ENABLED_LANGUAGES)


def set_supported_languages(portal):
    """Set supported languages to the same predictable set for all test layers."""
    language_tool = getToolByName(portal, "portal_languages")
    for lang in ENABLED_LANGUAGES:
        language_tool.addSupportedLanguage(lang)


def enable_request_language_negotiation(portal):
    """Enable request language negotiation during tests.

    This is so we can use the Accept-Language header to request translated
    pieces of content in different languages.
    """
    registry = getUtility(IRegistry)
    settings = registry.forInterface(ILanguageSchema, prefix="plone")
    settings.use_request_negotiation = True


class DateTimeFixture(Layer):
    def setUp(self):
        tz = "UTC"
        os.environ["TZ"] = tz
        time.tzset()

        # Patch DateTime's timezone for deterministic behavior.
        from DateTime import DateTime

        self.DT_orig_localZone = DateTime.localZone
        DateTime.localZone = lambda cls=None, ltm=None: tz

        from plone.dexterity import content

        content.FLOOR_DATE = DateTime(1970, 0)
        content.CEILING_DATE = DateTime(2500, 0)
        self._orig_content_zone = content._zone
        content._zone = tz

    def tearDown(self):
        if "TZ" in os.environ:
            del os.environ["TZ"]
        time.tzset()

        from DateTime import DateTime

        DateTime.localZone = self.DT_orig_localZone

        from plone.dexterity import content

        content._zone = self._orig_content_zone
        content.FLOOR_DATE = DateTime(1970, 0)
        content.CEILING_DATE = DateTime(2500, 0)


DATE_TIME_FIXTURE = DateTimeFixture()


class PloneRestApiDXLayer(PloneSandboxLayer):

    defaultBases = (DATE_TIME_FIXTURE, PLONE_APP_CONTENTTYPES_FIXTURE)

    def setUpZope(self, app, configurationContext):
        import plone.restapi

        xmlconfig.file("configure.zcml", plone.restapi, context=configurationContext)
        xmlconfig.file("testing.zcml", plone.restapi, context=configurationContext)

        self.loadZCML(package=collective.MockMailHost)
        z2.installProduct(app, "plone.restapi")

    def setUpPloneSite(self, portal):
        portal.acl_users.userFolderAddUser(
            SITE_OWNER_NAME, SITE_OWNER_PASSWORD, ["Manager"], []
        )
        login(portal, SITE_OWNER_NAME)
        setRoles(portal, TEST_USER_ID, ["Manager"])

        set_supported_languages(portal)

        applyProfile(portal, "plone.restapi:default")
        applyProfile(portal, "plone.restapi:testing")
        add_catalog_indexes(portal, DX_TYPES_INDEXES)
        set_available_languages()
        enable_request_language_negotiation(portal)
        quickInstallProduct(portal, "collective.MockMailHost")
        applyProfile(portal, "collective.MockMailHost:default")
        states = portal.portal_workflow["simple_publication_workflow"].states
        states["published"].title = "Published with accent é"  # noqa: E501


PLONE_RESTAPI_DX_FIXTURE = PloneRestApiDXLayer()
PLONE_RESTAPI_DX_INTEGRATION_TESTING = IntegrationTesting(
    bases=(PLONE_RESTAPI_DX_FIXTURE,), name="PloneRestApiDXLayer:Integration"
)
PLONE_RESTAPI_DX_FUNCTIONAL_TESTING = FunctionalTesting(
    bases=(PLONE_RESTAPI_DX_FIXTURE, z2.ZSERVER_FIXTURE),
    name="PloneRestApiDXLayer:Functional",
)


class PloneRestApiTestWorkflowsLayer(PloneSandboxLayer):

    defaultBases = (PLONE_RESTAPI_DX_FIXTURE,)

    def setUpPloneSite(self, portal):
        applyProfile(portal, "plone.restapi:testing-workflows")


PLONE_RESTAPI_WORKFLOWS_FIXTURE = PloneRestApiTestWorkflowsLayer()
PLONE_RESTAPI_WORKFLOWS_INTEGRATION_TESTING = IntegrationTesting(
    bases=(PLONE_RESTAPI_WORKFLOWS_FIXTURE,),
    name="PloneRestApiTestWorkflowsLayer:Integration",
)


class PloneRestApiDXPAMLayer(PloneSandboxLayer):

    defaultBases = (DATE_TIME_FIXTURE, PLONE_APP_CONTENTTYPES_FIXTURE)

    def setUpZope(self, app, configurationContext):
        import plone.restapi

        xmlconfig.file("configure.zcml", plone.restapi, context=configurationContext)
        xmlconfig.file("testing.zcml", plone.restapi, context=configurationContext)

        z2.installProduct(app, "plone.restapi")

    def setUpPloneSite(self, portal):
        portal.acl_users.userFolderAddUser(
            SITE_OWNER_NAME, SITE_OWNER_PASSWORD, ["Manager"], []
        )
        login(portal, SITE_OWNER_NAME)
        setRoles(portal, TEST_USER_ID, ["Manager"])

        set_supported_languages(portal)
        if portal.portal_setup.profileExists("plone.app.multilingual:default"):
            applyProfile(portal, "plone.app.multilingual:default")
        applyProfile(portal, "plone.restapi:default")
        applyProfile(portal, "plone.restapi:testing")
        add_catalog_indexes(portal, DX_TYPES_INDEXES)
        set_available_languages()
        enable_request_language_negotiation(portal)
        states = portal.portal_workflow["simple_publication_workflow"].states
        states["published"].title = "Published with accent é"  # noqa: E501


PLONE_RESTAPI_DX_PAM_FIXTURE = PloneRestApiDXPAMLayer()
PLONE_RESTAPI_DX_PAM_INTEGRATION_TESTING = IntegrationTesting(
    bases=(PLONE_RESTAPI_DX_PAM_FIXTURE,), name="PloneRestApiDXPAMLayer:Integration"
)
PLONE_RESTAPI_DX_PAM_FUNCTIONAL_TESTING = FunctionalTesting(
    bases=(PLONE_RESTAPI_DX_PAM_FIXTURE, z2.ZSERVER_FIXTURE),
    name="PloneRestApiDXPAMLayer:Functional",
)


class PloneRestApiDXIterateLayer(PloneSandboxLayer):

    defaultBases = (PLONEAPPITERATEDEX_FIXTURE,)

    def setUpZope(self, app, configurationContext):
        import plone.restapi

        xmlconfig.file("configure.zcml", plone.restapi, context=configurationContext)
        xmlconfig.file("testing.zcml", plone.restapi, context=configurationContext)

        z2.installProduct(app, "plone.restapi")


PLONE_RESTAPI_ITERATE_FIXTURE = PloneRestApiDXIterateLayer()
PLONE_RESTAPI_ITERATE_INTEGRATION_TESTING = IntegrationTesting(
    bases=(PLONE_RESTAPI_ITERATE_FIXTURE,),
    name="PloneRestApiDXIterateLayer:Integration",
)
PLONE_RESTAPI_ITERATE_FUNCTIONAL_TESTING = FunctionalTesting(
    bases=(PLONE_RESTAPI_ITERATE_FIXTURE, z2.ZSERVER_FIXTURE),
    name="PloneRestApiDXIterateLayer:Functional",
)


class PloneRestApIBlocksLayer(PloneSandboxLayer):

    defaultBases = (PLONE_RESTAPI_DX_FIXTURE,)

    def setUpPloneSite(self, portal):
        applyProfile(portal, "plone.restapi:blocks")


PLONE_RESTAPI_BLOCKS_FIXTURE = PloneRestApIBlocksLayer()
PLONE_RESTAPI_BLOCKS_INTEGRATION_TESTING = IntegrationTesting(
    bases=(PLONE_RESTAPI_BLOCKS_FIXTURE,), name="PloneRestApIBlocksLayer:Integration"
)
PLONE_RESTAPI_BLOCKS_FUNCTIONAL_TESTING = FunctionalTesting(
    bases=(PLONE_RESTAPI_BLOCKS_FIXTURE, z2.ZSERVER_FIXTURE),
    name="PloneRestApIBlocksLayer:Functional",
)


class RelativeSession(requests.Session):
    """A session that takes a base URL and makes requests relative to that
    base if their URL is relative (doesn't begin with a HTTP[S] scheme).
    """

    def __init__(self, base_url, test=None):
        """
        Capture the base URL.  Optionally also capture a test case for cleanup.

        Apparently, network sockets created by the `requests` library can remain open
        even after the full body of the response has been read, despite [the
        docs](https://docs.python-requests.org/en/latest/user/advanced/#body-content-workflow). In
        particular, this results in `ResourceWarning: unclosed <socket.socket ...>` leak
        warnings when running the tests.  If the `test` kwarg is passed, it will be used
        to register future cleanup calls to close this session and thus also the
        sockets. If passed, it must be an object with a `addCleanup(func)` callable
        attribute, such as instances of `unittest.TestCase`.
        """
        super().__init__()

        if not base_url.endswith("/"):
            base_url += "/"
        self.__base_url = base_url

        if hasattr(test, "addCleanup"):
            # Avoid `ResourceWarning: unclosed <socket.socket ...>`
            test.addCleanup(self.close)

    def request(self, method, url, **kwargs):
        if urlparse(url).scheme not in ("http", "https"):
            url = url.lstrip("/")
            url = urljoin(self.__base_url, url)
        try:
            return super().request(method, url, **kwargs)
        except ConnectionError:
            # On Jenkins we often get one ConnectionError in a seemingly
            # random test, mostly in test_documentation.py.
            # The server is still listening: the port is open.  We retry once.
            time.sleep(1)
            return super().request(method, url, **kwargs)


@implementer(IUUIDGenerator)
class StaticUUIDGenerator:
    """UUID generator that produces stable UUIDs for use in tests.

    Based on code from ftw.testing
    """

    def __init__(self, prefix):
        self.prefix = prefix[:26]
        self.counter = 0

    def __call__(self):
        self.counter += 1
        postfix = str(self.counter).rjust(32 - len(self.prefix), "0")
        return self.prefix + postfix


def register_static_uuid_utility(prefix):
    prefix = re.sub(r"[^a-zA-Z0-9\-_]", "", prefix)
    generator = StaticUUIDGenerator(prefix)
    getGlobalSiteManager().registerUtility(component=generator)



from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import SITE_OWNER_PASSWORD
from plone.restapi.permissions import UseRESTAPI
from plone.restapi.testing import PLONE_RESTAPI_DX_INTEGRATION_TESTING
from unittest import TestCase
from zExceptions import Unauthorized
from zope.event import notify
from ZPublisher.pubevents import PubStart


class TestLogin(TestCase):

    layer = PLONE_RESTAPI_DX_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]

    def traverse(self, path="/plone/@login", accept="application/json", method="POST"):
        request = self.layer["request"]
        request.environ["PATH_INFO"] = path
        request.environ["PATH_TRANSLATED"] = path
        request.environ["HTTP_ACCEPT"] = accept
        request.environ["REQUEST_METHOD"] = method
        notify(PubStart(request))
        return request.traverse(path)

    def test_login_without_pas_plugin_fails(self):
        self.portal.acl_users._delOb("jwt_auth")
        service = self.traverse()
        res = service.reply()
        self.assertIn("error", res)
        self.assertNotIn("token", res)

    def test_login_without_credentials_fails(self):
        service = self.traverse()
        res = service.reply()
        self.assertIn("error", res)
        self.assertNotIn("token", res)

    def test_login_with_invalid_credentials_fails(self):
        self.request["BODY"] = '{"login": "admin", "password": "admin"}'
        service = self.traverse()
        res = service.reply()
        self.assertIn("error", res)
        self.assertNotIn("token", res)

    def test_successful_login_returns_token(self):
        self.request["BODY"] = '{{"login": "{}", "password": "{}"}}'.format(
            SITE_OWNER_NAME,
            SITE_OWNER_PASSWORD,
        )
        service = self.traverse()
        res = service.reply()
        self.assertEqual(200, self.request.response.getStatus())
        self.assertIn("token", res)

    def test_invalid_token_returns_400(self):
        invalid_token = "abc123"
        self.request._auth = f"Bearer {invalid_token}"
        self.assertRaises(Unauthorized, self.traverse, path="/plone")

    def test_expired_token_returns_400(self):
        self.portal.acl_users.jwt_auth.store_tokens = True
        token = self.portal.acl_users.jwt_auth.create_token("admin", timeout=-60)
        self.request._auth = f"Bearer {token}"
        self.assertRaises(Unauthorized, self.traverse, path="/plone")

    def test_login_without_api_permission(self):
        self.portal.manage_permission(UseRESTAPI, roles=[])
        self.request["BODY"] = '{{"login": "{}", "password": "{}"}}'.format(
            SITE_OWNER_NAME,
            SITE_OWNER_PASSWORD,
        )
        service = self.traverse()
        res = service.render()
        self.assertIn("token", res)

    def test_login_with_zope_user_fails_without_pas_plugin(self):
        uf = self.layer["app"].acl_users
        uf.plugins.users.addUser("zopeuser", "zopeuser", "secret")
        if "jwt_auth" in uf:
            uf["jwt_auth"].manage_activateInterfaces([])
        self.request["BODY"] = '{"login": "zopeuser", "password": "secret"}'
        service = self.traverse()
        res = service.reply()
        self.assertIn("error", res)
        self.assertEqual(
            "JWT authentication plugin not installed.", res["error"]["message"]
        )
        self.assertNotIn("token", res)

    def test_login_with_zope_user(self):
        self.layer["app"].acl_users.plugins.users.addUser(
            "zopeuser", "zopeuser", "secret"
        )
        self.request["BODY"] = '{"login": "zopeuser", "password": "secret"}'
        service = self.traverse()
        res = service.reply()
        self.assertEqual(200, self.request.response.getStatus())
        self.assertIn("token", res)


class TestLogout(TestCase):

    layer = PLONE_RESTAPI_DX_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]

    def traverse(self, path="/plone/@logout", accept="application/json", method="POST"):
        request = self.layer["request"]
        request.environ["PATH_INFO"] = path
        request.environ["PATH_TRANSLATED"] = path
        request.environ["HTTP_ACCEPT"] = accept
        request.environ["REQUEST_METHOD"] = method
        notify(PubStart(request))
        return request.traverse(path)

    def test_logout_without_pas_plugin_fails(self):
        self.portal.acl_users._delOb("jwt_auth")
        service = self.traverse()
        res = service.reply()
        self.assertIn("error", res)

    def test_logout_with_not_stored_token_just_logouts_user(self):
        self.portal.acl_users.jwt_auth.store_tokens = False
        self.assertEqual(200, self.request.response.getStatus())

    def test_logout_with_without_credentials_fails(self):
        self.portal.acl_users.jwt_auth.store_tokens = True
        service = self.traverse()
        res = service.reply()
        self.assertEqual(400, self.request.response.getStatus())
        self.assertEqual("Unknown token", res["error"]["message"])

    def test_logout_with_invalid_token_fails(self):
        self.portal.acl_users.jwt_auth.store_tokens = True
        token = "my_invalid_token"
        self.request._auth = f"Bearer {token}"
        service = self.traverse()
        service.reply()
        self.assertEqual(400, self.request.response.getStatus())

    def test_logout_succeeds(self):
        self.portal.acl_users.jwt_auth.store_tokens = True
        token = self.portal.acl_users.jwt_auth.create_token("admin")
        self.request._auth = f"Bearer {token}"
        service = self.traverse()
        service.reply()
        self.assertEqual(200, self.request.response.getStatus())


class TestRenew(TestCase):

    layer = PLONE_RESTAPI_DX_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]

    def traverse(
        self, path="/plone/@login-renew", accept="application/json", method="POST"
    ):
        request = self.layer["request"]
        request.environ["PATH_INFO"] = path
        request.environ["PATH_TRANSLATED"] = path
        request.environ["HTTP_ACCEPT"] = accept
        request.environ["REQUEST_METHOD"] = method
        notify(PubStart(request))
        return request.traverse(path)

    def test_renew_without_pas_plugin_fails(self):
        self.portal.acl_users._delOb("jwt_auth")
        service = self.traverse()
        res = service.reply()
        self.assertIn("error", res)

    def test_renew_returns_token(self):
        self.portal.acl_users.jwt_auth.store_tokens = True
        token = self.portal.acl_users.jwt_auth.create_token("admin")
        self.request._auth = f"Bearer {token}"
        service = self.traverse()
        res = service.reply()
        self.assertIn("token", res)

    def test_renew_deletes_old_token(self):
        self.portal.acl_users.jwt_auth.store_tokens = True
        token = self.portal.acl_users.jwt_auth.create_token("admin")
        self.request._auth = f"Bearer {token}"
        service = self.traverse()
        res = service.reply()
        self.assertIn("token", res)
        self.assertEqual(1, len(self.portal.acl_users.jwt_auth._tokens["admin"]))

    def test_renew_fails_on_invalid_token(self):
        token = "this is an invalid token"
        self.request._auth = f"Bearer {token}"
        service = self.traverse()
        res = service.reply()
        self.assertEqual(service.request.response.status, 401)
        self.assertEqual(
            res["error"]["type"], "Invalid or expired authentication token"
        )



from setuptools import setup, find_packages
import sys

version = "8.22.0"

assert sys.version_info >= (3, 6, 0), "plone.restapi 8 requires Python 3.6.0+. Please downgrade to plone.restapi 7 for Python 2 and Plone 4.3/5.1."


def read(filename):
    with open(filename) as myfile:
        try:
            return myfile.read()
        except UnicodeDecodeError:
            # Happens on one Jenkins node on Python 3.6,
            # so maybe it happens for users too.
            pass
    # Opening and reading as text failed, so retry opening as bytes.
    with open(filename, "rb") as myfile:
        contents = myfile.read()
        return contents.decode("utf-8")


long_description = (
    read("README.rst")
    + "\n"
    + "Contributors\n"
    + "============\n"
    + "\n"
    + read("CONTRIBUTORS.rst")
    + "\n"
    + read("CHANGES.rst")
    + "\n"
)

TEST_REQUIRES = [
    "collective.MockMailHost",
    "plone.app.contenttypes",
    "plone.app.robotframework",
    "plone.app.testing [robot] >= 4.2.2",  # ROBOT_TEST_LEVEL added
    "plone.api",
    "requests",
    "plone.tiles",
    "mock",
]

setup(
    name="plone.restapi",
    version=version,
    description="plone.restapi is a RESTful hypermedia API for Plone.",
    long_description=long_description,
    # Get more strings from
    # https://pypi.org/classifiers/
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Framework :: Plone",
        "Framework :: Plone :: 5.2",
        "Framework :: Plone :: 6.0",
        "Framework :: Plone :: Core",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    keywords="plone rest restful hypermedia api json",
    author="Timo Stollenwerk (kitconcept GmbH)",
    author_email="tisto@plone.org",
    url="https://github.com/plone/plone.restapi/",
    license="gpl",
    packages=find_packages("src"),
    package_dir={"": "src"},
    namespace_packages=["plone"],
    python_requires=">=3.6.0",
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        "setuptools",
        "python-dateutil",
        "plone.behavior>=1.1",  # adds name to behavior directive
        "plone.rest >= 1.0a6",  # json renderer moved to plone.restapi
        "plone.schema >= 1.2.1",  # new/fixed json field
        "PyJWT",
        "pytz",
    ],
    extras_require={"test": TEST_REQUIRES},
    entry_points="""
      # -*- Entry points: -*-
      [z3c.autoinclude.plugin]
      target = plone
      """,
)



from plone.restapi.types.adapters import ListJsonSchemaProvider
from plone.restapi.types.interfaces import IJsonSchemaProvider
from z3c.relationfield.interfaces import IRelationList
from zope.component import adapter
from zope.component import getMultiAdapter
from zope.interface import implementer
from zope.interface import Interface


@adapter(IRelationList, Interface, Interface)
@implementer(IJsonSchemaProvider)
class ChoiceslessRelationListSchemaProvider(ListJsonSchemaProvider):
    def get_items(self):
        """Get items properties."""
        value_type_adapter = getMultiAdapter(
            (self.field.value_type, self.context, self.request), IJsonSchemaProvider
        )

        # Prevent rendering all choices.
        value_type_adapter.should_render_choices = False

        return value_type_adapter.get_schema()



from plone.restapi.controlpanels import IControlpanel
from plone.restapi.services import Service
from zExceptions import BadRequest
from zope.component import getAdapters
from zope.interface import implementer
from zope.publisher.interfaces import IPublishTraverse


@implementer(IPublishTraverse)
class ControlpanelsDelete(Service):
    def __init__(self, context, request):
        super().__init__(context, request)
        self.params = []

    def publishTraverse(self, request, name):
        self.params.append(name)
        return self

    def get_controlpanel_adapters(self):
        adapters = getAdapters((self.context, self.request), provided=IControlpanel)
        for name, panel in adapters:
            panel.__name__ = name
            yield name, panel

    def panel_by_name(self, name):
        panels = dict(self.get_controlpanel_adapters())
        return panels.get(name)

    def reply(self):
        if len(self.params) < 2:
            raise BadRequest("Can't delete Control Panel: %s" % self.params)

        panel = self.panel_by_name(self.params[0])
        panel.delete(self.params[1:])

        return self.reply_no_content()
